# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Initialization for Jupyter Notebooks.

This module performs several steps to initialize MSTICPy:
- imports a number of standard packages (e.g. pandas) into the notebook
- imports a number of modules and functions from msticpy
- checks the version of MSTICPy
- checks for presence and validates msticpyconfig.yaml

In the Azure ML and Azure Synapse environments, some additional
initialization and checks are performed.

One frequent problem in bootstrapping MSTICPy is obtaining or
creating a valid `msticpyconfig.yaml`. This is needed for many
configuration settings such as MS Sentinel workspaces, Threat
Intelligence (TI) providers, Azure authentication, Key Vault
settings and more.

If you are having trouble with this, you should check out the
following resources:

The basic structure of msticpyconfig.yaml

https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html

Using the msticpy settings editor to create or modify
msticpyconfig.yaml

https://msticpy.readthedocs.io/en/latest/getting_started/SettingsEditor.html

MS Sentinel and Azure ML resources:

Getting started notebook
https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/A%20Getting%20Started%20Guide%20For%20Azure%20Sentinel%20ML%20Notebooks.ipynb

Configuring your environment notebook
https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/ConfiguringNotebookEnvironment.ipynb

"""
import importlib
import io
import logging
import os
import sys
import traceback
import warnings
from contextlib import redirect_stdout
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple

import ipywidgets as widgets
import pandas as pd
from IPython import get_ipython
from IPython.core.interactiveshell import InteractiveShell
from IPython.display import HTML, display

try:
    import seaborn as sns
except ImportError:
    sns = None

from .._version import VERSION
from ..common.check_version import check_version
from ..common.exceptions import MsticpyException, MsticpyUserError
from ..common.pkg_config import _HOME_PATH
from ..common.settings import (
    current_config_path,
    get_config,
    refresh_config,
    validate_config,
)
from ..common.utility import (
    check_and_install_missing_packages,
    check_kwargs,
    is_ipython,
    md,
    search_for_file,
    unit_testing,
)
from .azure_ml_tools import check_aml_settings, is_in_aml, populate_config_to_mp_config
from .azure_synapse_tools import init_synapse, is_in_synapse
from .pivot import Pivot
from .user_config import load_user_defaults

__version__ = VERSION
__author__ = "Ian Hellen"


logger = logging.getLogger(__name__)

_IMPORT_ERR_MSSG = """
<h2><font color='red'>One or more missing packages detected</h2>
Please correct these by installing the required packages, restart
the kernel and re-run the notebook.</font>
<i>Package error: {err}</i><br>
"""

_IMPORT_MODULE_MSSG = """
<font color='red'>Error import module {module}</font>
"""

_MISSING_PKG_WARN = """
<h3><font color='orange'>Warning {package} is not installed or has an
incorrect version</h3></font>
"""

_HELP_URIS = (
    (
        '<li><a href="https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/'
        'A%20Getting%20Started%20Guide%20For%20Azure%20Sentinel%20ML%20Notebooks.ipynb"'
        'target="_blank" rel="noopener noreferrer">'
        "Getting Started (notebook)</a></li>"
    ),
    (
        '<li><a href="https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/'
        'ConfiguringNotebookEnvironment.ipynb"'
        'target="_blank" rel="noopener noreferrer">'
        "Configuring your Notebook environment (notebook)</a></li>"
    ),
    (
        '<li><a href="https://msticpy.readthedocs.io/en/latest/getting_started/'
        'msticpyconfig.html"'
        'target="_blank" rel="noopener noreferrer">'
        "Configuring MSTICPy settings (doc)</a></li>"
    ),
    (
        '<li><a href="https://msticpy.readthedocs.io/en/latest/getting_started/'
        'SettingsEditor.html"'
        'target="_blank" rel="noopener noreferrer">'
        "MSTICPy settings editor (doc)</a></li>"
    ),
    (
        '<li><a href="https://github.com/Azure/Azure-Sentinel-Notebooks/blob/'
        'master/TroubleShootingNotebooks.ipynb"'
        'target="_blank" rel="noopener noreferrer">'
        "Trouble-Shooting Notebooks (notebook)</a></li>"
    ),
)

_MISSING_MPCONFIG_ENV_ERR = f"""
<h3><font color='orange'>Warning: no <i>msticpyconfig.yaml</i> found</h3></font>
The MSTICPYCONFIG environment variable is set but does not point
to a valid file.<br>
Some functionality (such as Threat Intel lookups) will not function without
valid configuration settings.<br>
The following resources will help you set up your configuration:
<ul>{"".join(_HELP_URIS)}</ul>
<br>The first two of these are notebooks from the <b>Microsoft Sentinel
Notebooks</b> repo.
You can download and run these to set up your configuration in
another tab.
"""


_PANDAS_REQ_VERSION = (0, 25, 0)


def _get_verbosity_setting() -> Callable[[Optional[int]], int]:
    """Closure for holding trace setting."""
    _verbosity = 1

    def _verbose(verbosity: Optional[int] = None) -> int:
        nonlocal _verbosity
        if verbosity is not None:
            _verbosity = verbosity
        return _verbosity

    return _verbose


_VERBOSITY: Callable[[Optional[int]], int] = _get_verbosity_setting()

# pylint: disable=use-dict-literal
_NB_IMPORTS = [
    dict(pkg="pandas", alias="pd"),
    dict(pkg="IPython", tgt="get_ipython"),
    dict(pkg="IPython.display", tgt="display"),
    dict(pkg="IPython.display", tgt="HTML"),
    dict(pkg="IPython.display", tgt="Markdown"),
    # dict(pkg="ipywidgets", alias="widgets"),
    dict(pkg="pathlib", tgt="Path"),
    dict(pkg="numpy", alias="np"),
]
if sns is not None:
    _NB_IMPORTS.append(dict(pkg="seaborn", alias="sns"))

_MP_IMPORTS = [
    dict(pkg="msticpy"),
    dict(pkg="msticpy.data", tgt="QueryProvider"),
    # dict(pkg="msticpy.vis.foliummap", tgt="FoliumMap"),
    # dict(pkg="msticpy.context", tgt="TILookup"),
    # dict(pkg="msticpy.context", tgt="GeoLiteLookup"),
    # dict(pkg="msticpy.context", tgt="IPStackLookup"),
    # dict(pkg="msticpy.transform", tgt="IoCExtract"),
    dict(pkg="msticpy.common.utility", tgt="md"),
    dict(pkg="msticpy.common.utility", tgt="md_warn"),
    dict(pkg="msticpy.common.wsconfig", tgt="WorkspaceConfig"),
    dict(pkg="msticpy.init.pivot", tgt="Pivot"),
    dict(pkg="msticpy.datamodel", tgt="entities"),
    dict(pkg="msticpy.init", tgt="nbmagics"),
    # dict(pkg="msticpy.nbtools", tgt="SecurityAlert"),
    dict(pkg="msticpy.vis", tgt="mp_pandas_plot"),
    # dict(pkg="msticpy.vis", tgt="nbdisplay"),
    dict(pkg="msticpy.init", tgt="mp_pandas_accessors"),
    # dict(pkg="msticpy", tgt="nbwidgets"),
]

_MP_IMPORT_ALL: List[Dict[str, str]] = [
    dict(module_name="msticpy.datamodel.entities"),
]
# pylint: enable=use-dict-literal

_CONF_URI = (
    "https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html"
)

_AZNB_GUIDE = (
    "Please run the <i>Getting Started Guide for Azure Sentinel "
    + "ML Notebooks</i> notebook."
)

current_providers: Dict[str, Any] = {}  # pylint: disable=invalid-name

_SYNAPSE_KWARGS = ["identity_type", "storage_svc_name", "tenant_id", "cloud"]


# pylint: disable=too-many-statements, too-many-branches
def init_notebook(
    namespace: Optional[Dict[str, Any]] = None,
    def_imports: str = "all",
    additional_packages: List[str] = None,
    extra_imports: List[str] = None,
    **kwargs,
):
    """
    Initialize the notebook environment.

    Parameters
    ----------
    namespace : Dict[str, Any], optional
        Namespace (usually globals()) into which imports
        are to be populated.
        By default, it will use the ipython `user_global_ns`.
    def_imports : str, optional
        Import default packages. By default "all".
        Possible values are:
        - "all" - import all packages
        - "nb" - import common notebook packages
        - "msticpy" - import msticpy packages
        - "none" (or any other value) don't load any default packages.
    additional_packages : List[str], optional
        Additional packages to be pip installed,
        by default None.
        Packages are specified by name only or version
        specification (e.g. "pandas>=0.25")
    user_install : bool, optional
        Install packages in the "user" rather than system site-packages.
        Use this option if you cannot or do not want to update the system
        packages.
        You should usually avoid using this option with standard Conda environments.
    extra_imports : List[str], optional
        Additional import definitions, by default None.
        Imports are specified as up to 3 comma-delimited values
        in a string:
        "{source_pkg}, [{import_tgt}], [{alias}]"
        `source_pkg` is mandatory - equivalent to a simple "import xyz"
        statement.
        `{import_tgt}` specifies an object to import from the package
        equivalent to "from source_pkg import import_tgt"
        `alias` allows renaming of the imported object - equivalent to
        the "as alias" part of the import statement.
        If you want to provide just `source_pkg` and `alias` include
        an additional placeholder comma: e.g. "pandas, , pd"
    friendly_exceptions : Optional[bool]
        Setting this to True causes msticpy to hook the notebook
        exception handler. Any exceptions derived from MsticpyUserException
        are displayed but do not produce a stack trace, etc.
        Defaults to system/user settings if no value is supplied.
    verbose : Union[int, bool], optional
        Controls amount if status output, by default 1
        0 = No output
        1 or False = Brief output (default)
        2 or True = Detailed output
    verbosity : int, optional
        alias for `verbose`
    config : Optional[str]
        Use this path to load a msticpyconfig.yaml.
        Defaults are MSTICPYCONFIG env variable, home folder (~/.msticpy),
        current working directory.
    no_config_check : bool, optional
        Skip the check for valid configuration. Default is False.
    detect_env : Union[bool, List[str]], optional
        By default init_notebook tries to detect environments and makes
        additional configuration changes if it finds one of the supported
        notebook environments.
        Passing `False` for this parameter disables environment detection.
        Alternatively, you can pass a list of environments that you
        want to detect. Current supported environments are 'aml' (Azure
        Machine Learning) and 'synapse' (Azure Synapse).

    Raises
    ------
    MsticpyException :
        If extra_imports data format is incorrect.
        If package with required version check has no version
        information.

    Notes
    -----
    This module performs several steps to initialize MSTICPy:
    - imports a number of standard packages (e.g. pandas) into the notebook
    - imports a number of modules and functions from msticpy
    - checks the version of MSTICPy
    - checks for presence and validates msticpyconfig.yaml

    In the Azure ML and Azure Synapse environments, some additional
    initialization and checks are performed.

    One frequent problem in bootstrapping MSTICPy is obtaining or
    creating a valid `msticpyconfig.yaml`. This is needed for many
    configuration settings such as MS Sentinel workspaces, Threat
    Intelligence (TI) providers, Azure authentication, Key Vault
    settings and more.

    If you are having trouble with this, you should do the following:
    1. Run `init_notebook` again using the `verbosity=2` parameter
    This will print out additional status and debugging information

    2. Run mp.MpConfigEdit() to edit (or create) a msticpyconfig file.

    3. Check out the following resources:

    The basic structure of msticpyconfig.yaml

    https://msticpy.readthedocs.io/en/latest/getting_started/msticpyconfig.html

    Using the msticpy settings editor to create or modify
    msticpyconfig.yaml

    https://msticpy.readthedocs.io/en/latest/getting_started/SettingsEditor.html

    MS Sentinel and Azure ML resources:

    Getting started notebook
    https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/A%20Getting%20Started%20Guide%20For%20Azure%20Sentinel%20ML%20Notebooks.ipynb

    Configuring your environment notebook
    https://github.com/Azure/Azure-Sentinel-Notebooks/blob/master/ConfiguringNotebookEnvironment.ipynb

    """
    if namespace is None and get_ipython():
        namespace = get_ipython().user_global_ns
    else:
        namespace = namespace if (namespace is not None) else {}
    check_kwargs(
        kwargs,
        [
            "user_install",
            "friendly_exceptions",
            "no_config_check",
            "verbosity",
            "verbose",
            "config",
            *_SYNAPSE_KWARGS,
        ],
    )
    user_install: bool = kwargs.pop("user_install", False)
    friendly_exceptions: Optional[bool] = kwargs.pop("friendly_exceptions", None)
    no_config_check: bool = kwargs.pop("no_config_check", False)
    if "config" in kwargs:
        _use_custom_config(kwargs.pop("config", None))

    _set_verbosity(**kwargs)

    _pr_output("<hr><h4>Starting Notebook initialization...</h4>")
    logger.info("Starting Notebook initialization")
    # Check Azure ML environment
    if _detect_env("aml", **kwargs) and is_in_aml():
        try:
            check_aml_settings(*_get_aml_globals(namespace))
        except Exception as err:  # pylint: disable=broad-except
            logger.warning("Error initializing Azure ML environment: %s", err)
    else:
        # If not in AML check and print version status
        _check_msticpy_version()

    if _detect_env("synapse", **kwargs) and is_in_synapse():
        synapse_params = {
            key: val for key, val in kwargs.items() if key in _SYNAPSE_KWARGS
        }
        try:
            init_synapse(**synapse_params)
        except Exception as err:  # pylint: disable=broad-except
            logger.warning("Error initializing Synapse environment: %s", err)

    # Handle required packages and imports
    _pr_output("Processing imports....")
    imp_ok = _import_packages(
        namespace, def_imports, additional_packages, extra_imports, user_install
    )

    # Configuration check
    if no_config_check:
        conf_ok = True
    else:
        _pr_output("Checking configuration....")
        conf_ok = _get_or_create_config()

    # Notebook options
    _pr_output("Setting notebook options....")
    _set_nb_options(namespace)

    # Set friendly exceptions
    if friendly_exceptions is None:
        friendly_exceptions = get_config("msticpy.FriendlyExceptions", None)
    if friendly_exceptions:
        if _VERBOSITY() == 2:  # type: ignore
            _pr_output("Friendly exceptions enabled.")
        InteractiveShell.showtraceback = _hook_ipython_exceptions(  # type: ignore
            InteractiveShell.showtraceback
        )

    # load pivots
    _load_pivot_functions(namespace)

    # User defaults
    _load_user_defaults(namespace)

    # show any warnings
    _show_init_warnings(imp_ok, conf_ok)
    _pr_output("<h4>Notebook initialization complete</h4>")
    logger.info("Notebook initialization complete")


def _pr_output(*args):
    """Output to IPython display or print."""
    if not _VERBOSITY():
        return
    if is_ipython():
        display(HTML(" ".join([*args, "<br>"]).replace("\n", "<br>")))
    else:
        print(*args)


def _err_output(*args):
    """Output to IPython display or print - always output regardless of verbosity."""
    if is_ipython():
        display(HTML(" ".join([*args, "<br>"]).replace("\n", "<br>")))
        display(
            HTML(
                "For more info and options run:"
                "<pre>import msticpy as mp\nhelp(mp.nbinit)</pre>"
            )
        )
    else:
        print(*args)
        print(
            "\nFor more info and options run:",
            "\n    import msticpy as mp",
            "\n    help(mp.nbinit)",
        )


def _load_user_defaults(namespace):
    """Load user defaults, if defined."""
    global current_providers  # pylint: disable=global-statement, invalid-name
    stdout_cap = io.StringIO()
    with redirect_stdout(stdout_cap):
        _pr_output("Loading user defaults.")
        prov_dict = load_user_defaults()
        output = stdout_cap.getvalue()
        if output.strip():
            _pr_output(output)
            logger.info(output)
            logger.info("User default load failures: %s", output)

    if prov_dict:
        namespace.update(prov_dict)
        current_providers = prov_dict
        _pr_output("Auto-loaded components:", ", ".join(prov_dict.keys()))


def _load_pivot_functions(namespace):
    """Load pivot functions."""
    stdout_cap = io.StringIO()
    with redirect_stdout(stdout_cap):
        _pr_output("Loading pivots.")
        _load_pivots(namespace=namespace)
        output = stdout_cap.getvalue()
        if output.strip():
            _pr_output(output)
            logger.info("Pivot load failures: %s", output)


def _import_packages(
    namespace, def_imports, additional_packages, extra_imports, user_install
):
    """Import packages from default set or supplied as parameters."""
    stdout_cap = io.StringIO()
    with redirect_stdout(stdout_cap):
        imp_ok = _global_imports(
            namespace, additional_packages, user_install, extra_imports, def_imports
        )
        output = stdout_cap.getvalue()
        if output.strip():
            _pr_output(output)
            logger.info("Import failures: %s", output)
    return imp_ok


def _check_msticpy_version():
    """Check msticpy version."""
    stdout_cap = io.StringIO()
    with redirect_stdout(stdout_cap):
        check_version()
        output = stdout_cap.getvalue()
        if output.strip():
            _pr_output(output)
            logger.info("Check version failures: %s", output)


def _show_init_warnings(imp_ok, conf_ok):
    """Show any warnings from init_notebook."""
    if imp_ok and conf_ok:
        return True
    md("<font color='orange'><h3>Notebook setup completed with some warnings.</h3>")
    if not imp_ok:
        md("One or more libraries did not import successfully.")
        md(_AZNB_GUIDE)
    if not conf_ok:
        md("One or more configuration items were missing or set incorrectly.")
        md(
            _AZNB_GUIDE
            + f" and the <a href='{_CONF_URI}'>msticpy configuration guide</a>."
        )
    md("This notebook may still run but with reduced functionality.")
    return False


def _set_verbosity(**kwargs):
    """Set verbosity of output from boolean or int `verbose` param."""
    verbosity = 0
    verb_param = kwargs.pop("verbose", kwargs.pop("verbosity", 0))
    if isinstance(verb_param, bool):
        verbosity = 2 if verb_param else 0
    elif isinstance(verb_param, int):
        verbosity = min(2, max(0, verb_param))
    _VERBOSITY(verbosity)


def _detect_env(env_name: Literal["aml", "synapse"], **kwargs):
    """Return true if an environment should be detected."""
    detect_opt = kwargs.get("detect_env", True)
    if isinstance(detect_opt, bool):
        return detect_opt
    return env_name in detect_opt if isinstance(detect_opt, list) else True


def list_default_imports():
    """List the default imports for `init_notebook`."""
    for imp_group in (_NB_IMPORTS, _MP_IMPORTS):
        for imp_item in imp_group:
            if "tgt" in imp_item:
                import_line = f"from {imp_item['pkg']} import {imp_item['tgt']}"
            else:
                import_line = f"import {imp_item['pkg']}"
            if "alias" in imp_item:
                import_line += f" as {imp_item['alias']}"
            _pr_output(import_line)
    for imp_item in _MP_IMPORT_ALL:
        _pr_output(f"from {imp_item['module_name']} import *")
        logger.info("from %s import *", imp_item["module_name"])


def _extract_pkg_name(
    imp_pkg: Optional[Dict[str, str]] = None,
    pkg: str = None,
    tgt: str = None,
    alias: str = None,
) -> str:
    """Return string representation of package import."""
    if imp_pkg:
        pkg = imp_pkg.get("pkg")
        tgt = imp_pkg.get("tgt")
        alias = imp_pkg.get("alias")
    import_item = f"{pkg}.{tgt}" if tgt else pkg
    if alias:
        import_item = f"{alias} ({import_item})"
    return import_item  # type: ignore


PY_VER_VAR = "REQ_PYTHON_VER"
MP_VER_VAR = "REQ_MSTICPY_VER"
MP_EXTRAS = "REQ_MP_EXTRAS"


def _get_aml_globals(namespace: Dict[str, Any]):
    """Return global values if found."""
    py_ver = namespace.get(PY_VER_VAR, "3.8")
    mp_ver = namespace.get(MP_VER_VAR, __version__)
    extras = namespace.get(MP_EXTRAS)
    return py_ver, mp_ver, extras


def _global_imports(
    namespace: Dict[str, Any],
    additional_packages: List[str] = None,
    user_install: bool = False,
    extra_imports: List[str] = None,
    def_imports: str = "all",
):
    """Import packages from default set (defined statically)."""
    import_list = []
    imports, imports_all = _build_import_list(def_imports)

    for imp_pkg in imports:
        try:
            _imp_from_package(nm_spc=namespace, **imp_pkg)
            import_list.append(_extract_pkg_name(imp_pkg))
        except ImportError as imp_err:
            display(HTML(_IMPORT_ERR_MSSG.format(err=imp_err)))
            logger.exception("Import failure")
    for imp_pkg in imports_all:
        try:
            _imp_module_all(nm_spc=namespace, **imp_pkg)
            import_list.append(_extract_pkg_name(imp_pkg))
        except ImportError as imp_err:
            display(HTML(_IMPORT_ERR_MSSG.format(err=imp_err)))
            logger.exception("Import failure")
    _check_and_reload_pkg(namespace, pd, _PANDAS_REQ_VERSION, "pd")

    if additional_packages:
        pkg_success = check_and_install_missing_packages(
            additional_packages, user=user_install
        )
        if not pkg_success:
            _err_output("One or more packages failed to install.")
            _err_output(
                "Please re-run init_notebook() with the parameter user_install=True."
            )
        # We want to force import lib to see anything that we've
        # just installed.
        importlib.invalidate_caches()
    if extra_imports:
        import_list.extend(
            _import_extras(nm_spc=namespace, extra_imports=extra_imports)
        )

    imported_items = f"Imported:{', '.join(imp for imp in import_list if imp)}"
    _pr_output(imported_items)
    logger.info(imported_items)
    return True


def _build_import_list(
    def_imports: str,
) -> Tuple[List[Dict[str, str]], List[Dict[str, str]]]:
    imports = []
    imports_all = []
    if def_imports.casefold() in {"all", "nb"}:
        imports.extend(_NB_IMPORTS)
    if def_imports.casefold() in {"all", "msticpy"}:
        imports.extend(_MP_IMPORTS)
        imports_all.extend(_MP_IMPORT_ALL)
    return imports, imports_all


_AZ_SENT_ERRS = [
    "Missing or empty 'AzureSentinel' section",
    "Missing or empty 'Workspaces' key in 'AzureSentinel' section",
]


def _verify_no_azs_errors(errs):
    """Verify none of the Microsoft Sentinel errors appear in `errs`."""
    return all(az_err not in errs for az_err in _AZ_SENT_ERRS)


def _use_custom_config(config_file: str):
    """Use the config file supplied as a command line parameter."""
    if not config_file:
        return
    if not Path(config_file).is_file():
        raise ValueError(f"Configuration file {config_file} not found")
    os.environ["MSTICPYCONFIG"] = config_file
    refresh_config()


def _get_or_create_config() -> bool:
    # Cases
    # 0. Current config file exists -> return ok
    # 1. Env var set and mpconfig exists -> goto 4
    # 2. Env var set and mpconfig file not exists - warn and continue
    # 3. search_for_file finds mpconfig -> goto 4
    # 4. if file and check_file_contents -> return ok
    # 5. search_for_file(config.json)
    # 6. If aml user try to import config.json into mpconfig and save
    # 7. Error - no Microsoft Sentinel config
    if current_config_path() is not None:
        return True
    mp_path = os.environ.get("MSTICPYCONFIG")
    if mp_path and not Path(mp_path).is_file():
        _err_output(_MISSING_MPCONFIG_ENV_ERR)
        logger.warning("MSTICPYCONFIG set to %s but no file found", mp_path)
    if not mp_path or not Path(mp_path).is_file():
        mp_path = search_for_file("msticpyconfig.yaml", paths=[".", _HOME_PATH])

    if mp_path:
        logger.info("msticpyconfig found %s", mp_path)
        errs: List[str] = []
        try:
            std_out_cap = io.StringIO()
            with redirect_stdout(std_out_cap):
                errs, _ = validate_config(config_file=mp_path)
            if errs:
                _pr_output(std_out_cap.getvalue())
                for err in errs:
                    logger.warning("config validation error %s", err)
            if _verify_no_azs_errors(errs):
                # If the mpconfig has a Microsoft Sentinel config, return here
                return True
        # pylint: disable=broad-except
        except Exception as err:
            errs.append(f"Exception while checking configuration:\n{err}")
            _pr_output(f"Exception while checking configuration:\n{type(err)} - {err}")
            _pr_output("\n".join(traceback.format_tb(err.__traceback__)))
            _pr_output("Please report this to msticpy@microsoft.com")
        # pylint: enable=broad-except

    _pr_output("Could not find msticpyconfig.yaml in standard search.")
    logger.warning("Could not find msticpyconfig.yaml in standard search.")
    if is_in_aml():
        _pr_output(
            "AML environment detected.",
            "Attempting to import settings from config.json to msticpyconfig.yaml.",
        )
        status = populate_config_to_mp_config(mp_path)
        if status:
            _pr_output(status)
            return True
    _pr_output("No valid configuration for Microsoft Sentinel found.")
    return False


def _set_nb_options(namespace):
    namespace["WIDGET_DEFAULTS"] = {
        "layout": widgets.Layout(width="95%"),
        "style": {"description_width": "initial"},
    }

    warnings.filterwarnings("ignore", category=DeprecationWarning)
    if sns:
        sns.set()
    pd.set_option("display.max_rows", 100)
    pd.set_option("display.max_columns", 50)
    pd.set_option("display.max_colwidth", 100)

    os.environ["KQLMAGIC_LOAD_MODE"] = "silent"
    # Kqlmagic config will use AZ CLI login if available
    kql_config = os.environ.get("KQLMAGIC_CONFIGURATION", "")
    if "try_azcli_login" not in kql_config:
        kql_config = ";".join([kql_config, "try_azcli_login=True"])
        os.environ["KQLMAGIC_CONFIGURATION"] = kql_config


def _load_pivots(namespace):
    """Load pivot functions."""
    # pylint: disable=no-member
    if not Pivot.current():
        pivot = Pivot()
        pivot.reload_pivots()
        namespace["pivot"] = pivot
        # pylint: disable=import-outside-toplevel, cyclic-import
        import msticpy

        setattr(msticpy, "pivot", pivot)


def _import_extras(nm_spc: Dict[str, Any], extra_imports: List[str]):
    added_imports = []
    if isinstance(extra_imports, str):
        extra_imports = [extra_imports]
    params: List[Optional[str]]
    for imp_spec in extra_imports:
        params = [None, None, None]
        for idx, param in enumerate(imp_spec.split(",")):
            params[idx] = param.strip() or None

        if params[0] is None:
            raise MsticpyException(
                f"First parameter in extra_imports is mandatory: {imp_spec}"
            )
        try:
            _imp_from_package(
                nm_spc=nm_spc, pkg=params[0], tgt=params[1], alias=params[2]
            )
            added_imports.append(
                _extract_pkg_name(pkg=params[0], tgt=params[1], alias=params[2])
            )
        except ImportError as imp_err:
            display(HTML(_IMPORT_ERR_MSSG.format(err=imp_err)))
            logger.exception("Import failure")
    return added_imports


def _imp_module(nm_spc: Dict[str, Any], module_name: str, alias: str = None):
    """Import named module and assign to global alias."""
    try:
        mod = importlib.import_module(module_name)
    except ImportError:
        _err_output(_IMPORT_MODULE_MSSG.format(module=module_name))
        return None
    if alias:
        nm_spc[alias] = mod
    else:
        nm_spc[module_name] = mod
    if _VERBOSITY() == 2:  # type: ignore
        _pr_output(f"{module_name} imported (alias={alias})")
    return mod


def _imp_module_all(nm_spc: Dict[str, Any], module_name):
    """Import all from named module add to globals."""
    try:
        imported_mod = importlib.import_module(module_name)
    except ImportError:
        _err_output(_IMPORT_MODULE_MSSG.format(module=module_name))
        return
    for item in dir(imported_mod):
        if item.startswith("_"):
            continue
        nm_spc[item] = getattr(imported_mod, item)
    if _VERBOSITY() == 2:  # type: ignore
        _pr_output(f"All items imported from {module_name}")


def _imp_from_package(
    nm_spc: Dict[str, Any], pkg: str, tgt: str = None, alias: str = None
):
    """Import object or submodule from `pkg`."""
    if not tgt:
        return _imp_module(nm_spc=nm_spc, module_name=pkg, alias=alias)
    try:
        # target could be a module
        obj = importlib.import_module(f".{tgt}", pkg)
    except ImportError:
        # if not, it must be an attribute (class, func, etc.)
        try:
            mod = importlib.import_module(pkg)
        except ImportError:
            _err_output(_IMPORT_MODULE_MSSG.format(module=pkg))
            return None
        obj = getattr(mod, tgt)
    if alias:
        nm_spc[alias] = obj
    else:
        nm_spc[tgt] = obj
    if _VERBOSITY() == 2:  # type: ignore
        _pr_output(f"{tgt} imported from {pkg} (alias={alias})")
    return obj


def _check_and_reload_pkg(
    nm_spc: Dict[str, Any], pkg: Any, req_version: Tuple[int, ...], alias: str = None
):
    """Check package version matches required version and reload."""
    warn_mssg = []
    pkg_name = pkg.__name__
    if not hasattr(pkg, "__version__"):
        raise MsticpyException(f"Package {pkg_name} has no version data.")
    pkg_version = tuple(int(v) for v in pkg.__version__.split("."))
    if pkg_version < req_version:
        _err_output(_MISSING_PKG_WARN.format(package=pkg_name))
        # sourcery skip: swap-if-expression
        resp = (
            input("Install the package now? (y/n)") if not unit_testing() else "y"
        )  # nosec
        if resp.casefold().startswith("y"):
            warn_mssg.append(f"{pkg_name} was installed or upgraded.")
            pip_ver = ".".join(str(elem) for elem in req_version)
            pkg_spec = f"{pkg_name}>={pip_ver}"
            check_and_install_missing_packages(required_packages=[pkg_spec], user=True)

            if pkg_name in sys.modules:
                importlib.reload(pkg)
            else:
                _imp_module(nm_spc, pkg_name, alias=alias)
    if _VERBOSITY() == 2:  # type: ignore
        _pr_output(f"{pkg_name} imported version {pkg.__version__}")
    return warn_mssg


def _hook_ipython_exceptions(func):
    """Hooks the `func` and bypasses it if exception is MsticpyUserException."""
    # if already wrapped, don't do it again
    if hasattr(InteractiveShell.showtraceback, "__wrapped__"):
        return InteractiveShell.showtraceback

    @wraps(func)
    def showtraceback(*args, **kwargs):
        """Replace IPython showtraceback."""
        # extract exception type, value and traceback
        e_type, exception, _ = sys.exc_info()
        if e_type is not None and issubclass(e_type, MsticpyUserError):
            exception.display_exception()
            return None
        # otherwise run the original hook
        return func(*args, **kwargs)

    return showtraceback


def reset_ipython_exception_handler():
    """Remove MSTICPy custom exception handler."""
    if hasattr(InteractiveShell.showtraceback, "__wrapped__"):
        InteractiveShell.showtraceback = InteractiveShell.showtraceback.__wrapped__



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Settings provider for secrets."""
from typing import Any, Set

import keyring
from keyring.errors import KeyringError, KeyringLocked

from .._version import VERSION
from ..common.utility import export

__version__ = VERSION
__author__ = "Ian Hellen"


@export
class KeyringClient:
    """Keyring client wrapper."""

    def __init__(self, name: str = "key-cache", debug: bool = False):
        """
        Initialize the keyring client.

        Parameters
        ----------
        name : str, optional
            Name of the credential group, by default "key-cache"
        debug : bool, optional
            Output debug info, by default False

        """
        self.debug = debug
        self.keyring = name
        self._secret_names: Set[str] = set()

    def __getitem__(self, key: str):
        """Get key name."""
        cred = self.get_secret(key)
        if cred:
            return cred
        raise KeyError

    def get_secret(self, secret_name: str) -> Any:
        """
        Retrieve a secret from the keyring.

        Parameters
        ----------
        secret_name : str
            Secret name.

        Returns
        -------
        Any
            Secret value.

        """
        secret = None
        if self.debug:
            print(f"Fetching {secret_name} from keyring")
        try:
            secret = keyring.get_password(self.keyring, secret_name)
        except (KeyringError, KeyringLocked):
            if self.debug:
                print(
                    "Keyring error retrieving credentials",
                    f"for {secret_name} from keyring {self.keyring}",
                )
        if not secret and self.debug:
            print("No credentials", f"for {secret_name} from keyring {self.keyring}")
        return secret

    def set_secret(self, secret_name: str, secret_value: Any):
        """
        Set a secret in the keyring group.

        Parameters
        ----------
        secret_name : str
            Name of the secret
        secret_value : Any
            Secret value

        """
        if self.debug:
            print(f"Saving {secret_name} to keyring {self.keyring}")
        self._secret_names.add(secret_name)
        keyring.set_password(self.keyring, secret_name, secret_value)

    def delete_secret(self, secret_name: str):
        """
        Delete a secret in the keyring group.

        Parameters
        ----------
        secret_name : str
            Name of the secret

        """
        if self.debug:
            print(f"Deleting {secret_name} from keyring {self.keyring}")
        self._secret_names.remove(secret_name)
        keyring.delete_password(self.keyring, secret_name)

    @staticmethod
    def is_keyring_available() -> bool:
        """
        Test if valid keyring backend is available.

        Returns
        -------
        bool
            True if Keyring has a usable backend, False if not.

        """
        return keyring.get_keyring() != keyring.backends.fail.Keyring



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Functions to support investigation of a domain or url.

Includes functions to conduct common investigation steps when dealing
with a domain or url, such as getting a screenshot or validating the TLD.

"""
from __future__ import annotations

import datetime as dt
import json
import logging
import ssl
import time
from dataclasses import asdict
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable
from urllib.error import HTTPError, URLError

import httpx
import pandas as pd
import tldextract
from cryptography import x509

# CodeQL [SM02167] Compatibility requirement for SSL abuse list
from cryptography.hazmat.primitives.hashes import SHA1  # CodeQL [SM02167] Compatibility
from dns.exception import DNSException
from dns.resolver import Resolver
from IPython import display
from ipywidgets import IntProgress
from typing_extensions import Self
from urllib3.exceptions import LocationParseError
from urllib3.util import parse_url

from .._version import VERSION
from ..common.exceptions import MsticpyUserConfigError
from ..common.settings import get_config, get_http_timeout
from ..common.utility import export, mp_ua_header

if TYPE_CHECKING:
    from cryptography.x509 import Certificate
    from dns.resolver import Answer
    from tldextract.tldextract import ExtractResult
__version__ = VERSION
__author__ = "Pete Bryan"
logger: logging.Logger = logging.getLogger(__name__)


@export
def screenshot(  # pylint: disable=too-many-locals
    url: str,
    api_key: str | None = None,
    *,
    sleep: float = 0.05,
    max_progress: int = 100,
) -> httpx.Response:
    """
    Get a screenshot of a url with Browshot.

    Parameters
    ----------
    url : str
        The url a screenshot is wanted for.
    api_key : str (optional)
        Browshot API key. If not set msticpyconfig checked for this.
    sleep: int (optional)
        Time to sleep between calls. Defaults to 0.05 seconds
    max_progress: int (optional)
        Set the maximum value for the progress bar. Defaults to 100.

    Returns
    -------
    image_data: httpx.Response
        The final screenshot request response data.

    """
    # Get Browshot API key from kwargs or config
    if api_key is not None:
        bs_api_key: str | None = api_key
    else:
        bs_conf: dict[str, Any] = get_config(
            "DataProviders.Browshot",
            {},
        ) or get_config(
            "Browshot",
            {},
        )
        bs_api_key = None
        if bs_conf is not None:
            bs_api_key = bs_conf.get("Args", {}).get("AuthKey")

    if bs_api_key is None:
        err_msg: str = (
            "No configuration found for Browshot\n"
            "Please add a section to msticpyconfig.yaml:\n"
            "DataProviders:\n"
            "  Browshot:\n"
            "    Args:\n"
            "      AuthKey: {your_auth_key}"
        )
        raise MsticpyUserConfigError(
            err_msg,
            title="Browshot configuration not found",
            browshot_uri=("Get an API key for Browshot", "https://api.browshot.com/"),
        )

    # Request screenshot from Browshot and get request ID
    id_string: str = (
        f"https://api.browshot.com/api/v1/screenshot/create?url={url}/"
        f"&instance_id=26&size=screen&cache=0&key={bs_api_key}"
    )
    id_data: httpx.Response = httpx.get(
        id_string,
        timeout=get_http_timeout(),
        headers=mp_ua_header(),
    )
    bs_id: str = json.loads(id_data.content)["id"]
    status_string: str = (
        f"https://api.browshot.com/api/v1/screenshot/info?id={bs_id}&key={bs_api_key}"
    )
    image_string: str = (
        f"https://api.browshot.com/api/v1/screenshot/thumbnail?id={bs_id}"
        f"&zoom=50&key={bs_api_key}"
    )
    # Wait until the screenshot is ready and keep user updated with progress
    logger.info("Getting screenshot")
    progress = IntProgress(min=0, max=max_progress)
    display.display(progress)
    ready = False
    while not ready and progress.value < max_progress:
        progress.value += 1
        status_data: httpx.Response = httpx.get(
            status_string,
            timeout=get_http_timeout(),
            headers=mp_ua_header(),
        )
        status: str = json.loads(status_data.content)["status"]
        if status == "finished":
            ready = True
        else:
            time.sleep(sleep)
    progress.value = max_progress

    # Once ready or timed out get the screenshot
    image_data: httpx.Response = httpx.get(image_string, timeout=get_http_timeout())

    if not image_data.is_success:
        logger.warning(
            "There was a problem with the request, please check the status code for details",
        )

    return image_data


# Backward compat with dnspython 1.x
# If v2.x installed use non-deprecated "resolve" method
# otherwise use "query"
_dns_resolver = Resolver()
if hasattr(_dns_resolver, "resolve"):
    _dns_resolve: Callable[..., Answer] = _dns_resolver.resolve
else:
    _dns_resolve = _dns_resolver.query


@export
class DomainValidator:
    """Assess a domain's validity."""

    _ssl_abuse_list: pd.DataFrame = pd.DataFrame()

    @classmethod
    def _check_and_load_abuselist(cls: type[Self]) -> None:
        """Pull IANA TLD list and save to internal attribute."""
        if cls._ssl_abuse_list is None or cls._ssl_abuse_list.empty:
            cls._ssl_abuse_list = cls._get_ssl_abuselist()

    @property
    def ssl_abuse_list(self: Self) -> pd.DataFrame:
        """
        Return the class SSL Blacklist.

        Returns
        -------
        pd.DataFrame
            SSL Blacklist

        """
        self._check_and_load_abuselist()
        return self._ssl_abuse_list

    @staticmethod
    def validate_tld(url_domain: str) -> bool:
        """
        Validate if a domain's TLD is valid.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid public TLD, False if not.

        """
        extract_result: ExtractResult = tldextract.extract(url_domain.lower())
        return bool(extract_result.suffix)

    @staticmethod
    def is_resolvable(url_domain: str) -> bool:
        """
        Validate if a domain or URL be be resolved to an IP address.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        result:
            True if valid resolvable, False if not.

        """
        try:
            _dns_resolve(url_domain, "A")
        except DNSException:
            return False
        return True

    def in_abuse_list(self: Self, url_domain: str) -> tuple[bool, Certificate | None]:
        """
        Validate if a domain or URL's SSL cert the abuse.ch SSL Abuse List.

        Parameters
        ----------
        url_domain : str
            The url or domain to validate.

        Returns
        -------
        Tuple[bool, Optional[Certificate]]:
            True if valid in the list, False if not.
            Certificate - the certificate loaded from the domain.

        """
        try:
            cert: str = ssl.get_server_certificate((url_domain, 443))
            x509_cert: Certificate = x509.load_pem_x509_certificate(
                cert.encode("ascii"),
            )
            cert_sha1: bytes = x509_cert.fingerprint(
                SHA1()
            )  # noqa: S303  # CodeQL [SM02167] Compatibility requirement for SSL abuse list
            result = bool(
                self.ssl_abuse_list["SHA1"].str.contains(cert_sha1.hex()).any(),
            )
        except Exception:  # pylint: disable=broad-except
            return False, None
        return result, x509_cert

    @classmethod
    def _get_ssl_abuselist(cls: type[Self]) -> pd.DataFrame:
        """Download and load abuse.ch SSL Abuse List."""
        try:
            ssl_ab_list: pd.DataFrame = pd.read_csv(
                "https://sslbl.abuse.ch/blacklist/sslblacklist.csv",
                skiprows=8,
            )
        except (ConnectionError, HTTPError, URLError):
            ssl_ab_list = pd.DataFrame({"SHA1": []})

        return ssl_ab_list


def dns_components(domain: str) -> dict:
    """
    Return components of domain as dict.

    Parameters
    ----------
    domain : str
        The domain to extract.

    Returns
    -------
    dict:
        Returns subdomain and TLD components from a domain.

    """
    result: ExtractResult = tldextract.extract(domain.lower())
    if isinstance(result, tuple) and hasattr(result, "_asdict"):
        return result._asdict()
    return asdict(result)


def url_components(url: str) -> dict[str, str]:
    """Return parsed Url components as dict."""
    try:
        return parse_url(url)._asdict()
    except LocationParseError:
        return {}


@export
def dns_resolve(url_domain: str, rec_type: str = "A") -> dict[str, Any]:
    """
    Validate if a domain or URL be be resolved to an IP address.

    Parameters
    ----------
    url_domain : str
        The url or domain to validate.
    rec_type : str
        The DNS record type to query, by default "A"

    Returns
    -------
    Dict[str, Any]:
        Resolver result as dictionary.

    """
    domain: str | None = parse_url(url_domain).host
    if not domain:
        err_msg: str = f"Failed to parse url: {url_domain}"
        raise ValueError(err_msg)
    try:
        return _resolve_resp_to_dict(_dns_resolve(domain, rdtype=rec_type))
    except DNSException as err:
        return {
            "qname": domain,
            "rdtype": rec_type,
            "response": str(err),
        }


@export
def dns_resolve_df(url_domain: str, rec_type: str = "A") -> pd.DataFrame:
    """
    Validate if a domain or URL be be resolved to an IP address.

    Parameters
    ----------
    url_domain : str
        The url or domain to validate.
    rec_type : str
        The DNS record type to query, by default "A"

    Returns
    -------
    pd.DataFrame:
        Resolver result as dataframe with individual resolution
        results as separate rows.

    """
    results = pd.DataFrame([dns_resolve(url_domain, rec_type)])
    if "rrset" in results.columns:
        return results.explode(column="rrset")
    return results


@export
def ip_rev_resolve(ip_address: str) -> dict[str, Any]:
    """
    Reverse lookup for IP Address.

    Parameters
    ----------
    ip_address : str
        The IP address to query.

    Returns
    -------
    Dict[str, Any]:
        Resolver result as dictionary.

    """
    try:
        return _resolve_resp_to_dict(_dns_resolve(ip_address, raise_on_no_answer=True))
    except DNSException as err:
        return {
            "qname": ip_address,
            "rdtype": "PTR",
            "response": str(err),
        }


@export
def ip_rev_resolve_df(ip_address: str) -> pd.DataFrame:
    """
    Reverse lookup for IP Address.

    Parameters
    ----------
    ip_address : str
        The IP address to query.

    Returns
    -------
    pd.DataFrame:
        Resolver result as dataframe with individual resolution
        results as separate rows.

    """
    results = pd.DataFrame([ip_rev_resolve(ip_address)])
    if "rrset" in results.columns:
        return results.explode(column="rrset")
    return results


@export
def _resolve_resp_to_dict(resolver_resp: Answer) -> dict[str, Any]:
    """Return Dns Python resolver response to dict."""
    rdtype: str = (
        resolver_resp.rdtype.name
        if isinstance(resolver_resp.rdtype, Enum)
        else str(resolver_resp.rdtype)
    )
    rdclass: str = (
        resolver_resp.rdclass.name
        if isinstance(resolver_resp.rdclass, Enum)
        else str(resolver_resp.rdclass)
    )

    result: dict[str, Any] = {
        "qname": str(resolver_resp.qname),
        "rdtype": rdtype,
        "rdclass": rdclass,
        "response": str(resolver_resp.response),
        "nameserver": getattr(resolver_resp, "nameserver", None),
        "port": getattr(resolver_resp, "port", None),
        "canonical_name": str(resolver_resp.canonical_name),
        "expiration": dt.datetime.fromtimestamp(
            resolver_resp.expiration,
            tz=dt.timezone.utc,
        ),
    }
    if resolver_resp.rrset:
        result["rrset"] = [str(res) for res in resolver_resp.rrset]
    return result



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for SecurityAlert class."""
import html
import re
from collections import Counter
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from deprecated.sphinx import deprecated

from .._version import VERSION
from ..common.utility import escape_windows_path, export
from ..data.core.query_defns import DataEnvironment, QueryParamProvider
from ..datamodel.entities import Account, Entity, Host, OSFamily, Process

__version__ = VERSION
__author__ = "Ian Hellen"

_ID_PROPERTIES: List[str] = [
    "AzSubscriptionId",
    "AzResourceId",
    "WorkspaceId",
    "AgentId",
    "TenantId",
    "SourceComputerId",
    "ResourceId",
    "WorkspaceSubscriptionId",
    "WorkspaceResourceGroup",
    "ProviderAlertId",
    "SystemAlertId",
    "ResourceId",
]


# pylint: disable=too-many-public-methods
@deprecated(reason="Replaced by Alert entity in datamodel", version="1.7.0")
@export
class SecurityBase(QueryParamProvider):
    """
    Security Base Class for alerts and events.

    Instantiates a security event or alert from a pandas Series
    and provides convenience access methods to retrieve properties.
    """

    def __init__(self, src_row: pd.Series = None):
        """Instantiate a security alert from a pandas Series."""
        self._source_data: pd.Series = (
            src_row if src_row is not None else pd.Series(dtype="object")
        )
        self._custom_query_params: Dict[str, Any] = {}
        self._entities: List[Entity] = []

        # Extract and cache alert ID properties
        self._ids: Dict[str, str] = {}
        if self._source_data is not None:
            for id_property in _ID_PROPERTIES:
                if id_property in self._source_data:
                    self._ids[id_property] = self._source_data[id_property]

        self.path_separator = "\\"
        self.os_family = "Windows"

    # Dunder methods
    def __getitem__(self, key):
        """Allow property get using dictionary key syntax."""
        if key in self.__dict__:
            return self.__dict__[key]
        if key in self._source_data:
            return self._source_data[key]
        raise KeyError

    def __setitem__(self, key, value):
        """Allow property set using dictionary key syntax."""
        self.__dict__[key] = value

    def __contains__(self, key):
        """Allow property 'in' test."""
        # get attribute from source dictionary if not defined
        return key in self.__dict__ or key in self._source_data

    def __getattr__(self, name):
        """Return the value of the named property 'name'."""
        if self._source_data is not None and name in self._source_data:
            return self._source_data[name]
        raise AttributeError(f"{name} is not a valid attribute.")

    def __str__(self):
        """Return string representation of object properties."""
        str_props = [f"{prop}: {val}" for prop, val in self._source_data.items()]

        if self.entities:
            str_entities = [str(ent).replace("\n", ", ") for ent in self.entities]
            str_props += str_entities
        return "\n".join(str_props)

    def __repr__(self) -> str:
        """Return repr of item."""
        if self.properties:
            params = ", ".join(
                [f"{name}={val}" for name, val in self.properties.items()]
            )
            if len(params) > 80:
                params = params[:80] + "..."
            return f"{self.__class__.__name__}({params})"
        return f"{self.__class__.__name__}()"

    def _repr_html_(self) -> str:
        """Display in IPython."""
        return self.to_html()

    # def __getstate__(self):
    #     """Return dictionary of state for serialization/pickling."""
    #     state_dict = {}
    #     state_dict['source_data'] = (self._source_data)
    #     state_dict['Entities'] = self._entities
    #     state_dict['path_separator'] = self.path_separator
    #     state_dict['os_family'] = self.os_family
    #     return state_dict

    # def __setstate__(self, state):
    #     """Set state from dictionary for deserialization/unpickling."""
    #     self._source_data = state['source_data']
    #     self._entities = list(state['Entities'])
    #     self.path_separator = state['path_separator']
    #     self.os_family = state['os_family']

    # Properties
    @property
    def entities(self) -> List[Entity]:
        """
        Return a list of the Alert or Event entities.

        Returns
        -------
        List[Entity]
            List of the Alert or Event entities.

        """
        return self._entities

    @property
    def properties(self) -> Dict[str, Any]:
        """
        Return a dictionary of the Alert or Event properties.

        Returns
        -------
        Dict[str, Any]
            dictionary of the Alert or Event properties.

        """
        return self._source_data.to_dict()

    @property
    def hostname(self) -> str:
        """Return the Hostname (not FQDN) of the host associated with the alert."""
        return self.primary_host.HostName if self.primary_host is not None else None

    @property
    def computer(self) -> Optional[str]:
        """
        Return the Computer name of the host associated with the alert.

        (host FQDN, if available)
        """
        return self.primary_host.computer if self.primary_host is not None else None

    @property
    def ids(self) -> Dict[str, str]:
        """Return a collection of Identity properties for the alert."""
        return self._ids

    @property
    def is_in_workspace(self) -> bool:
        """Return True if the alert has a Log Analytics WorkspaceID."""
        return "WorkspaceId" in self._ids and "AgentId" in self._ids

    @property
    def is_in_log_analytics(self) -> bool:
        """Return True if the alert originates from a Log Analytics Workspace host."""
        return "TenantId" in self._ids

    @property
    def is_in_azure_sub(self) -> bool:
        """Return True if the alert originates from an Azure Security Center host."""
        if (
            "AzSubscriptionId" not in self._ids
            and "AzResourceId" not in self._ids
            and "ResourceId" in self._ids
            and self._ids["ResourceId"]
        ):
            self._ids["AzResourceId"] = self._id["ResourceId"]
            res = self._get_subscription_from_resource(self._id["ResourceId"])
            if res:
                self._ids["AzSubscriptionId"] = res

        return "AzSubscriptionId" in self._ids and "AzResourceId" in self._ids

    @property
    def primary_host(self) -> Optional[Union[Host, Entity]]:
        """
        Return the primary host entity (if any) associated with this object.

        Returns
        -------
        Optional[Host]
            primary host entity (if any)

        """
        hosts = self.get_entities_of_type("host")
        if hosts:
            return hosts[0]
        return None

    @property
    def primary_process(self) -> Optional[Union[Process, Entity]]:
        """
        Return the primary process entity (if any) associated with this object.

        Returns
        -------
        Optional[Process]
            primary process entity (if any)

        """
        procs = self.get_entities_of_type("process")
        if not procs:
            return None
        if len(procs) == 1:
            return procs[0]

        # find the first process that has a parent process property
        procs_with_parent = [
            proc
            for proc in procs
            if "ParentProcess" in proc
            and proc["ParentProcess"]
            and proc.ParentProcess.ProcessId in {proc.ProcessId for proc in procs}
        ]
        return procs_with_parent[0] if procs_with_parent else procs[0]

    @property
    def primary_account(self) -> Optional[Union[Process, Entity]]:
        """
        Return the primary account entity (if any) associated with this object.

        Returns
        -------
        Optional[Process]
            primary account entity (if any)

        """
        accts = self.get_entities_of_type("account")
        return accts[0] if accts else None

    @property
    def query_params(self) -> Dict[str, Any]:
        """
        Query parameters derived from alert.

        Returns
        -------
        Dict[str, Any]
            Dictionary of parameter names/values

        """
        try:
            host_name = self.primary_host.fqdn if self.primary_host else None
            proc_name = (
                self.primary_process.ImageFile.FullPath
                if self.primary_process and self.primary_process.ImageFile
                else None
            )
            acct_name = self.primary_account.Name if self.primary_account else None
            path_separator = self.path_separator
            if self.data_family == OSFamily.Windows:
                proc_name = escape_windows_path(proc_name)
                path_separator = escape_windows_path(self.path_separator)

            dyn_query_params = {
                "subscription_filter": self.subscription_filter(),
                "host_filter_eq": self.host_filter(operator="=="),
                "host_filter_neq": self.host_filter(operator="!="),
                "host_name": host_name,
                "account_name": acct_name,
                "process_name": proc_name,
                "logon_session_id": self.get_logon_id(),
                "process_id": (
                    self.primary_process.ProcessId if self.primary_process else None
                ),
                "path_separator": path_separator,
                "data_family": self.data_family,
                "data_environment": self.data_environment,
            }

            dyn_query_params.update(self._custom_query_params)
            return dyn_query_params
        except AttributeError:
            return {}

    @property
    def data_family(self) -> OSFamily:
        """Return the data family of the alert for subsequent queries."""
        if self.os_family == "Linux":
            return OSFamily.Linux
        if self.os_family == "Windows":
            return OSFamily.Windows
        raise ValueError("Unknown Data family.")

    @property
    def data_environment(self) -> DataEnvironment:
        """Return the data environment of the alert for subsequent queries."""
        if self.is_in_log_analytics:
            return DataEnvironment.LogAnalytics
        return DataEnvironment.Kusto

    @property
    def origin_time(self) -> datetime:
        """Return the datetime of event."""
        return self.TimeGenerated

    def get_logon_id(self, account: Account = None) -> Optional[Union[str, int]]:
        """
        Get the logon Id for the alert or the account, if supplied.

        If `account` is not supplied, return the logon id
        of the first host-logon-session or account entity.

        Parameters
        ----------
        account : Account, optional
            Account objec to use (the default is None)

        Returns
        -------
        Optional[Union[str, int]]
            The logon Id for primary account

        """
        for session in [
            e
            for e in self.entities
            if e["Type"] in ["host-logon-session", "hostlogonsession"]
        ]:
            if account is None or session["Account"] == account:
                return session["SessionId"]
        if account is None:
            for acct in [
                e for e in self.entities if e["Type"] == "account" and "LogonId" in e
            ]:
                return acct["LogonId"]
        elif "LogonId" in account:
            return account["LogonId"]
        return None

    def subscription_filter(self, operator="=="):
        """Return a KQL subscription filter clause derived from the alert properties."""
        if self.is_in_log_analytics:
            return "true"
        if self.is_in_azure_sub:
            return (
                f"AzureResourceSubscriptionId {operator} "
                f"'{self._ids['AzSubscriptionId']}'"
            )
        if self.is_in_workspace:
            return f"WorkspaceId {operator} '{self._ids['WorkspaceId']}'"

        # Otherwise we default to including everything
        return "true"

    def host_filter(self, operator="=="):
        """
        Return a KQL host filter clause derived from the alert properties.

            :param operator='==': the operator to use in the filter clause.
                '==' and '!=' typically.
        """
        if self.primary_host:
            case_insens_op = "=~" if operator == "==" else "!~"
            return f"Computer {case_insens_op} '{self.primary_host.computer}'"

        if (
            self.is_in_log_analytics
            and "SourceComputerId" in self._ids
            and self._ids["SourceComputerId"]
        ):
            return f"SourceComputerId {operator} '{self._ids['SourceComputerId']}'"
        if (
            self.is_in_azure_sub
            and "AzureResourceId" in self._ids
            and self._ids["AzResourceId"]
        ):
            return f"AzureResourceId {operator} '{self._ids['AzResourceId']}'"
        if self.is_in_workspace and "AgendId" in self._ids and self._ids["AgentId"]:
            return f"AgentId {operator} '{self._ids['AgentId']}'"
        return None

    def get_entities_of_type(self, entity_type: str) -> List[Entity]:
        """
        Return entity collection for a give entity type.

        Parameters
        ----------
        entity_type : str, optional
            The entity type.

        Returns
        -------
        List[Entity]
            The entities matching `entity_type`.

        """
        class_type = Entity.ENTITY_NAME_MAP.get(entity_type, None)
        return [
            p
            for p in self.entities
            if p["Type"] == entity_type or class_type and isinstance(p, class_type)
        ]

    def get_all_entities(self) -> pd.DataFrame:
        """
        Return a DataFrame of the Alert or Event entities.

        Returns
        -------
        DataFrame
            Pandas DataFrame of the Alert or Event entities.

        """
        entity = []
        ent_type = []
        for item in self.entities:
            if "Address" in item:
                entity.append(item["Address"])
                ent_type.append(item["Type"])
            elif "Url" in item:
                entity.append(item["Url"])
                ent_type.append(item["Type"])
            elif "HostName" in item:
                entity.append(item["HostName"])
                ent_type.append(item["Type"])
            elif "Entity" in item:
                entity.append(item["Entity"])
                ent_type.append(item["Type"])
            elif item["Type"] == "account":
                entity.append(item["Name"])
                ent_type.append(item["Type"])

        return pd.DataFrame({"Entity": entity, "Type": ent_type})

    def to_html(self, show_entities: bool = False) -> str:
        """Return the item as HTML string."""
        html_doc = pd.DataFrame(self._source_data).to_html()

        if self._source_data is not None and "ExtendedProperties" in self._source_data:
            ext_prop_title = "<br/><h3>ExtendedProperties:</h3>"
            ext_prop_html = pd.DataFrame(
                pd.Series(self._source_data["ExtendedProperties"])
            ).to_html()
            html_doc = html_doc + ext_prop_title + ext_prop_html

        if show_entities and self.entities:
            entity_title = "<br/><h3>Entities:</h3><br/>"
            entity_html = "<br/>".join(
                [self._format_entity(ent) for ent in self.entities]
            )
            html_doc = html_doc + entity_title + entity_html
        else:
            e_counts = Counter([ent["Type"] for ent in self.entities])
            e_counts_str = ", ".join([f"{e}: {c}" for e, c in e_counts.items()])
            html_doc = html_doc + f"<h3>Entity counts: </h3>{e_counts_str}"
        return html_doc

    @staticmethod
    def _format_entity(entity):
        str_entity = str(entity)
        if str_entity:
            str_entity = str_entity.replace("\n", ", ")
        return html.escape(str_entity)

    def _find_os_family(self):
        """Return OSFamily and path separator to use from entities or file paths."""
        self.path_separator = "\\"
        self.os_family = "Windows"

        # Use OSFamily if any entities have this property set
        os_family_entities = [e for e in self.entities if "OSFamily" in e]
        if os_family_entities:
            for os_entity in os_family_entities:
                if os_entity["OSFamily"] == "Linux":
                    self.os_family = "Linux"
                    self.path_separator = "/"
                    break
        else:
            # Otherwise try to infer from the file paths
            files = [e for e in self.entities if e["Type"] == "file"]
            if files:
                for file in files:
                    if "Directory" in file and "/" in file["Directory"]:
                        self.os_family = "Linux"
                        self.path_separator = "/"
                        break
            else:
                for proc in [
                    e
                    for e in self.entities
                    if e["Type"] == "process" and "ImageFile" in e
                ]:
                    file = proc["ImageFile"]
                    if "Directory" in file and "/" in file["Directory"]:
                        self.os_family = "Linux"
                        self.path_separator = "/"
                        break

    @staticmethod
    def _get_subscription_from_resource(resource_id) -> Optional[str]:
        """Extract subscription Id from resource string."""
        sub_regex = r"^/subscriptions/([^/]+)/"
        sub_ids = re.findall(sub_regex, resource_id, re.RegexFlag.I)
        if sub_ids:
            return sub_ids[0]

        return None



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Utility Modules related to AI agents used in MSTICpy."""

import os
from typing import Callable, Dict, List, Union

from azure.identity import DefaultAzureCredential, get_bearer_token_provider

from ..common.exceptions import MsticpyUserConfigError
from ..common.pkg_config import get_config

ConfigItem = Dict[str, Union[str, Callable]]
ConfigList = List[ConfigItem]
Config = Dict[str, Union[str, float, ConfigList]]

token_provider = get_bearer_token_provider(
    DefaultAzureCredential(), "https://cognitiveservices.azure.com/.default"
)


def inject_token_provider_callable(
    config: ConfigItem,
) -> ConfigItem:
    """Replace autogen configuration `azure_ad_token_provider` with a token provider callable.

    Parameters
    ----------
    config : ConfigItem
        Autogen LLM configuration.

    Returns
    -------
    ConfigItem
        Autogen LLM configuration with the token provider callable.
    """
    if "azure_ad_token_provider" in config:
        config["azure_ad_token_provider"] = token_provider

    return config


def inject_environment_variable(config: ConfigItem) -> ConfigItem:
    """Replace autogen configuration `api_key` with the value of an environment variable.

    Parameters
    ----------
    config : ConfigItem
        Autogen LLM configuration.

    Returns
    -------
    ConfigItem
        Autogen LLM configuration with the environment variable value.
    """
    if "api_key" in config:
        api_key = os.environ.get(str(config["api_key"]), None)
        if not api_key:
            raise MsticpyUserConfigError(
                f"Environment variable {config['api_key']} specified, but not found!"
            )
        config["api_key"] = api_key

    return config


def get_autogen_config_from_msticpyconfig() -> Config:
    """Get Autogen configuration from msticpyconfig.yaml.

    See `https://microsoft.github.io/autogen/docs/topics/llm_configuration`
    for more information on the structure of the LLM configuration object.

    Please note that a configuration list is required, rather than a single configuration.

    Returns
    -------
    Config
        Autogen configuration.

    Raises
    ------
    MsticpyUserConfigError
        Autogen settings not found in msticpyconfig.yaml configuration
    MsticpyUserConfigError
        Autogen.config_list setting not found in msticpyconfig.yaml configuration
    """
    autogen_config = get_config("Autogen", None)
    if not autogen_config:
        raise MsticpyUserConfigError(
            "Autogen settings not found in msticpyconfig.yaml configuration!"
        )

    if "config_list" not in autogen_config:
        raise MsticpyUserConfigError(
            "Autogen.config_list setting not found in msticpyconfig.yaml configuration!"
        )

    injectors = [
        inject_token_provider_callable,
        inject_environment_variable,
    ]

    config_list = []
    for config in autogen_config["config_list"]:
        for injector in injectors:
            config = injector(config)
        config_list.append(config)

    return {
        **autogen_config,
        "config_list": config_list,
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Data driver base class."""
import abc
from abc import ABC
from collections import defaultdict
from typing import Any, Dict, Iterable, Optional, Set, Tuple, Union

import pandas as pd

from ..._version import VERSION
from ...common.exceptions import MsticpyNotConnectedError
from ...common.pkg_config import get_http_timeout
from ...common.provider_settings import ProviderSettings, get_provider_settings
from ..core.query_defns import DataEnvironment
from ..core.query_source import QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"


class DriverProps:
    """Defined driver properties."""

    PUBLIC_ATTRS = "public_attribs"
    FORMATTERS = "formatters"
    USE_QUERY_PATHS = "use_query_paths"
    HAS_DRIVER_QUERIES = "has_driver_queries"
    EFFECTIVE_ENV = "effective_environment"
    SUPPORTS_THREADING = "supports_threading"
    SUPPORTS_ASYNC = "supports_async"
    MAX_PARALLEL = "max_parallel"
    FILTER_ON_CONNECT = "filter_queries_on_connect"

    PROPERTY_TYPES: Dict[str, Any] = {
        PUBLIC_ATTRS: dict,
        FORMATTERS: dict,
        USE_QUERY_PATHS: bool,
        HAS_DRIVER_QUERIES: bool,
        EFFECTIVE_ENV: (str, DataEnvironment),
        SUPPORTS_THREADING: bool,
        SUPPORTS_ASYNC: bool,
        MAX_PARALLEL: int,
        FILTER_ON_CONNECT: bool,
    }

    @classmethod
    def defaults(cls):
        """Return default values for driver properties."""
        return {
            cls.PUBLIC_ATTRS: {},
            cls.FORMATTERS: {},
            cls.USE_QUERY_PATHS: True,
            cls.HAS_DRIVER_QUERIES: False,
            cls.EFFECTIVE_ENV: None,
            cls.SUPPORTS_THREADING: False,
            cls.SUPPORTS_ASYNC: False,
            cls.MAX_PARALLEL: 4,
            cls.FILTER_ON_CONNECT: False,
        }

    @classmethod
    def valid_type(cls, property_name: str, value: Any) -> bool:
        """Return expected property type."""
        if property_name not in cls.PROPERTY_TYPES:
            return True
        return isinstance(value, cls.PROPERTY_TYPES[property_name])


# pylint: disable=too-many-instance-attributes
class DriverBase(ABC):
    """Base class for data providers."""

    def __init__(self, **kwargs):
        """Initialize new instance."""
        self._kwargs = kwargs
        self._loaded = False
        self._connected = False
        self.current_connection = None
        # self.public_attribs: Dict[str, Any] = {}
        # self.formatters: Dict[str, Callable] = {}
        # self.use_query_paths = True
        # self.has_driver_queries = False
        self._previous_connection = False
        self.data_environment = kwargs.get("data_environment")
        self._query_filter: Dict[str, Set[str]] = defaultdict(set)
        self._instance: Optional[str] = None

        self.properties = DriverProps.defaults()
        self.set_driver_property(
            name=DriverProps.EFFECTIVE_ENV,
            value=(
                self.data_environment.name
                if isinstance(self.data_environment, DataEnvironment)
                else self.data_environment or ""
            ),
        )
        self.set_driver_property(DriverProps.SUPPORTS_THREADING, False)
        self.set_driver_property(DriverProps.MAX_PARALLEL, kwargs.get("max_threads", 4))

    def __getattr__(self, attrib):
        """Return item from the properties dictionary as an attribute."""
        if attrib in self.properties:
            return self.properties[attrib]
        raise AttributeError(f"{self.__class__.__name__} has no attribute '{attrib}'")

    @property
    def loaded(self) -> bool:
        """
        Return true if the provider is loaded.

        Returns
        -------
        bool
            True if the provider is loaded.

        Notes
        -----
        This is not relevant for some providers.

        """
        return self._loaded

    @property
    def connected(self) -> bool:
        """
        Return true if at least one connection has been made.

        Returns
        -------
        bool
            True if a successful connection has been made.

        Notes
        -----
        This does not guarantee that the last data source
        connection was successful. It is a best effort to track
        whether the provider has made at least one successful
        authentication.

        """
        return self._connected

    @property
    def instance(self) -> Optional[str]:
        """
        Return instance name, if one is set.

        Returns
        -------
        Optional[str]
            The name of driver instance or None if
            the driver does not support multiple instances

        """
        return self._instance

    @property
    def schema(self) -> Dict[str, Dict]:
        """
        Return current data schema of connection.

        Returns
        -------
        Dict[str, Dict]
            Data schema of current connection.

        """
        return {}

    @abc.abstractmethod
    def connect(self, connection_str: Optional[str] = None, **kwargs):
        """
        Connect to data source.

        Parameters
        ----------
        connection_str : Optional[str]
            Connect to a data source

        """

    @abc.abstractmethod
    def query(
        self, query: str, query_source: Optional[QuerySource] = None, **kwargs
    ) -> Union[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame of results.

        Parameters
        ----------
        query : str
            The query to execute
        query_source : QuerySource
            The query definition object

        Other Parameters
        ----------------
        kwargs :
            Are passed to the underlying provider query method,
            if supported.

        Returns
        -------
        Union[pd.DataFrame, Any]
            A DataFrame (if successful) or
            the underlying provider result if an error.

        """

    @abc.abstractmethod
    def query_with_results(self, query: str, **kwargs) -> Tuple[pd.DataFrame, Any]:
        """
        Execute query string and return DataFrame plus native results.

        Parameters
        ----------
        query : str
            The query to execute

        Returns
        -------
        Tuple[pd.DataFrame,Any]
            A DataFrame and native results.

        """

    @property
    def service_queries(self) -> Tuple[Dict[str, str], str]:
        """
        Return queries retrieved from the service after connecting.

        Returns
        -------
        Tuple[Dict[str, str], str]
            Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        return {}, ""

    @property
    def driver_queries(self) -> Iterable[Dict[str, Any]]:
        """
        Return queries retrieved from the service after connecting.

        Returns
        -------
        List[Dict[str, str]]
            List of Dictionary of query_name, query_text.
            Name of container to add queries to.

        """
        return [{}]

    @property
    def query_attach_spec(self) -> Dict[str, Set[str]]:
        """Parameters that determine whether a query is relevant for the driver."""
        return self._query_filter

    def add_query_filter(self, name: str, query_filter: Union[str, Iterable]):
        """Add an expression to the query attach filter."""
        allowed_names = {"data_environments", "data_families", "data_sources"}
        if name not in allowed_names:
            raise ValueError(
                f"'name' {name} must be one of:",
                ", ".join(f"'{name}'" for name in allowed_names),
            )
        if isinstance(query_filter, str):
            self._query_filter[name].add(query_filter)
        else:
            self._query_filter[name].update(query_filter)

    def set_driver_property(self, name: str, value: Any):
        """Set an item in driver properties."""
        if not DriverProps.valid_type(name, value):
            raise TypeError(
                f"Property '{name}' is not the correct type.",
                f"Expected: '{DriverProps.PROPERTY_TYPES[name]}'.",
            )
        self.properties[name] = value

    def get_driver_property(self, name: str) -> Any:
        """Return value or KeyError from driver properties."""
        return self.properties[name]

    def query_usable(self, query_source: QuerySource) -> bool:
        """Return True if query should be exposed for this driver."""
        del query_source
        return True

    # Read values from configuration
    @staticmethod
    def _get_config_settings(prov_name) -> Dict[Any, Any]:
        """Get config from msticpyconfig."""
        data_provs = get_provider_settings(config_section="DataProviders")
        splunk_settings: Optional[ProviderSettings] = data_provs.get(prov_name)
        return getattr(splunk_settings, "args", {})

    @staticmethod
    def _create_not_connected_err(prov_name):
        return MsticpyNotConnectedError(
            "Please run the connect() method before running this method.",
            title=f"not connected to {prov_name}.",
            help_uri="https://msticpy.readthedocs.io/en/latest/DataProviders.html",
        )

    @staticmethod
    def get_http_timeout(**kwargs):
        """Get http timeout from settings or kwargs."""
        return get_http_timeout(**kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Deprecated location for TI Providers."""

from ..._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
IPQualityScore Provider.

This provider offers contextual lookup services and fraud scoring for IP addresses.
https://www.ipqualityscore.com/
"""
from __future__ import annotations

from typing import Any, ClassVar

from typing_extensions import Self

from ..._version import VERSION
from ..http_provider import APILookupParams
from .ti_http_provider import HttpTIProvider
from .ti_provider_base import ResultSeverity

__version__ = VERSION
__author__ = "Pete Bryan"


class IPQualityScore(HttpTIProvider):
    """IP Quality Score Lookup."""

    _BASE_URL = "https://www.ipqualityscore.com/api/json"

    _QUERIES: ClassVar[dict[str, APILookupParams]] = {
        # Supported API Types
        "ipv4": APILookupParams(
            path="/ip/{AuthKey}/{observable}",
        ),
    }

    _REQUIRED_PARAMS: ClassVar[list[str]] = ["AuthKey"]

    MEDIUM_SEVERITY: ClassVar[int] = 50
    HIGH_SEVERITY: ClassVar[int] = 80

    def parse_results(self: Self, response: dict) -> tuple[bool, ResultSeverity, Any]:
        """
        Return the details of the response.

        Parameters
        ----------
        response : Dict
            The returned data response

        Returns
        -------
        Tuple[bool, ResultSeverity, Any]
            bool = positive or negative hit
            ResultSeverity = enumeration of severity
            Object with match details

        """
        if self._failed_response(response) or not isinstance(
            response["RawResult"],
            dict,
        ):
            return False, ResultSeverity.information, "Not found."
        result = True
        result_dict: dict[str, Any] = {}
        result_dict.update(
            {
                "FraudScore": response["RawResult"].get("fraud_score"),
                "ISP": response["RawResult"].get("ISP"),
                "ASN": response["RawResult"].get("ASN"),
                "CountryOrRegion": response["RawResult"].get("country_code"),
                "Region": response["RawResult"].get("city"),
                "City": response["RawResult"].get("region"),
                "Organization": response["RawResult"].get("organization"),
                "Latitude": response["RawResult"].get("latitude"),
                "Longitude": response["RawResult"].get("longitude"),
                "IsMobile": response["RawResult"].get("mobile"),
                "IsProxy": response["RawResult"].get("proxy"),
                "IsTor": response["RawResult"].get("active_tor"),
                "IsVPN": response["RawResult"].get("active_vpn"),
                "IsBot": response["RawResult"].get("bot_status"),
                "AbuseStatus": response["RawResult"].get("recent_abuse"),
            },
        )

        severity = ResultSeverity.information
        if (
            response["RawResult"]["fraud_score"] > self.MEDIUM_SEVERITY
            and response["RawResult"]["fraud_score"] < self.HIGH_SEVERITY
        ):
            severity = ResultSeverity.warning
        elif response["RawResult"]["fraud_score"] >= self.HIGH_SEVERITY:
            severity = ResultSeverity.high
        return result, severity, result_dict



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Module for MSTICpy documentation utilities and retrieval agent configuration.

Includes functions to find documentation files and to set up retrieval
agents that assist security analysts by answering questions based on MSTICpy documentation.
"""

import sys
from pathlib import Path
from typing import List, Optional

from autogen.agentchat.chat import ChatResult
from autogen.agentchat.contrib.retrieve_assistant_agent import RetrieveAssistantAgent
from autogen.agentchat.contrib.retrieve_user_proxy_agent import RetrieveUserProxyAgent

from .._version import VERSION
from ..common.exceptions import MsticpyUserConfigError
from .config_utils import get_autogen_config_from_msticpyconfig

if sys.version_info < (3, 9):
    import importlib_resources as pkg_resources
else:
    import importlib.resources as pkg_resources


def find_rst_files() -> List[str]:
    """
    Find all .rst files in the docs/source directory of 'msticpy' package.

    Returns
    -------
    list of str
        List of paths to .rst files in the docs/source directory.
    """
    # Get the path to the docs/source directory of the package
    docs_path = Path(pkg_resources.files("msticpy")).parent / "docs" / "source"

    # Find all .rst files in the docs/source directory
    rst_files = list(str(fp) for fp in docs_path.rglob("*.rst"))

    return rst_files


def get_retrieval_assistant_agent(system_message: str = "") -> RetrieveAssistantAgent:
    """
    Create and return a RetrieveAssistantAgent.

    Parameters
    ----------
    system_message : str, optional
        Custom system message for the assistant.

    Returns
    -------
    RetrieveAssistantAgent
        Configured RetrieveAssistantAgent instance.
    """
    if not system_message:
        system_message = (
            "You are a helpful assistant to security analysts using MSTICpy."
        )
    return RetrieveAssistantAgent(
        name="assistant",
        system_message=system_message,
        llm_config=get_autogen_config_from_msticpyconfig(),
    )


def get_retrieval_user_proxy_agent(
    customized_prompt: Optional[str] = None,
) -> RetrieveUserProxyAgent:
    """
    Create and return a RetrieveUserProxyAgent.

    Parameters
    ----------
    customized_prompt : Optional[str], optional
        Custom prompt for the assistant agent, by default None.

    Returns
    -------
    RetrieveUserProxyAgent
        Configured RetrieveUserProxyAgent instance.

    Raises
    ------
    MsticpyUserConfigError
        Autogen settings not found in msticpyconfig.yaml configuration
    """
    rst_files = find_rst_files()
    autogen_config = get_autogen_config_from_msticpyconfig()

    default_model = None
    if "config_list" in autogen_config and isinstance(
        autogen_config["config_list"], list
    ):
        if autogen_config["config_list"]:
            default_config = autogen_config["config_list"][0]
            if "model" in default_config:
                default_model = default_config["model"]

    if not default_model:
        raise MsticpyUserConfigError(
            "Could not find a valid default Autogen model in msticpyconfig.yaml configuration!"
        )

    return RetrieveUserProxyAgent(
        name="ragproxyagent",
        human_input_mode="NEVER",
        max_consecutive_auto_reply=1,
        is_termination_msg=lambda x: True,
        retrieve_config={
            "task": "default",
            "docs_path": rst_files,
            "chunk_token_size": 2000,
            "customized_prompt": customized_prompt,
            "model": default_model,
            "vector_db": "chroma",
            "collection_name": f"MSTICpy_Docs_{VERSION}",
            "get_or_create": True,
        },
        code_execution_config=False,
    )


def ask_question(
    assistant_agent: RetrieveAssistantAgent,
    user_proxy_agent: RetrieveUserProxyAgent,
    question: str,
) -> ChatResult:
    """
    Ask a question using the assistant and user proxy agents.

    Parameters
    ----------
    assistant_agent : RetrieveAssistantAgent
        The assistant agent to use.
    user_proxy_agent : RetrieveUserProxyAgent
        The user proxy agent to use.
    question : str
        The question to ask.

    Returns
    -------
    ChatResult
        The result of the chat interaction.
    """
    assistant_agent.reset()
    return user_proxy_agent.initiate_chat(
        assistant_agent, message=user_proxy_agent.message_generator, problem=question
    )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Data sub-package.

This sub-package contains data-access related modules and sub-packages.

It is split into the following sub-packages:

- azure - Azure and MS Sentinel API related modules
- browsers - notebook interactive data browsers for various components
  such as Threat Intel, Data Queries
- context - context provider modules such as TI providers, geoip providers
- core - core data query provider components
- drivers - query provider interface drivers for external data sources.
- queries - queries for core data providers
- uploaders - loaders for some data services.

"""
from .._version import VERSION

# from ..common.exceptions import MsticpyImportExtraError
from ..lazy_importer import lazy_import

__version__ = VERSION

_LAZY_IMPORTS = {
    "msticpy.data.core.data_providers.QueryProvider",
    "msticpy.data.core.query_defns.DataEnvironment",
    "msticpy.data.core.query_defns.DataFamily",
}

module, __getattr__, __dir__ = lazy_import(__name__, _LAZY_IMPORTS)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Uses the Azure Python SDK to collect and return details related to Azure."""
from __future__ import annotations

import datetime
import logging
from dataclasses import asdict, dataclass, field
from importlib.metadata import version
from typing import TYPE_CHECKING, Any, Callable, Iterable

import numpy as np
import pandas as pd
from packaging.version import Version, parse
from typing_extensions import Self

from ..._version import VERSION
from ...auth.azure_auth import (
    AzCredentials,
    AzureCloudConfig,
    az_connect,
    fallback_devicecode_creds,
)
from ...auth.azure_auth_core import only_interactive_cred
from ...common.exceptions import (
    MsticpyAzureConfigError,
    MsticpyImportExtraError,
    MsticpyNotConnectedError,
    MsticpyResourceError,
)

try:
    from azure.common.exceptions import CloudError
    from azure.core.exceptions import ClientAuthenticationError
    from azure.mgmt.network import NetworkManagementClient
    from azure.mgmt.resource import ResourceManagementClient
    from azure.mgmt.resource.subscriptions import SubscriptionClient

    if parse(version("azure.mgmt.monitor")) > Version("1.0.1"):
        # Try new version but keep backward compat with 1.0.1
        from azure.mgmt.monitor import MonitorManagementClient
    else:
        from azure.mgmt.monitor import (  # type: ignore[attr-defined, no-redef]
            MonitorClient as MonitorManagementClient,
        )
    from azure.mgmt.compute import ComputeManagementClient

    if TYPE_CHECKING:
        from azure.mgmt.compute.models import VirtualMachineInstanceView
        from azure.mgmt.network.models import NetworkInterface
        from azure.mgmt.subscription.models import Subscription
except ImportError as imp_err:
    error_msg: str = (
        "Cannot use this feature without these azure packages installed:\n"
        "azure.mgmt.network\n"
        "azure.mgmt.resource\n"
        "azure.mgmt.monitor\n"
        "azure.mgmt.compute\n"
    )
    raise MsticpyImportExtraError(
        error_msg,
        title="Error importing azure module",
        extra="azure",
    ) from imp_err

__version__ = VERSION
__author__ = "Pete Bryan"

# pylint:disable=too-many-lines

logger: logging.Logger = logging.getLogger(__name__)

_CLIENT_MAPPING: dict[
    str,
    type[
        SubscriptionClient
        | ResourceManagementClient
        | NetworkManagementClient
        | MonitorManagementClient
        | ComputeManagementClient
    ],
] = {
    "sub_client": SubscriptionClient,
    "resource_client": ResourceManagementClient,
    "network_client": NetworkManagementClient,
    "monitoring_client": MonitorManagementClient,
    "compute_client": ComputeManagementClient,
}


@dataclass
class Items:  # pylint:disable=too-many-instance-attributes
    """attr class to build resource details dictionary."""

    resource_id: str | None = None
    name: str | None = None
    resource_type: str | None = None
    location: str | None = None
    tags: Any = None
    plan: Any = None
    properties: Any = None
    kind: str | None = None
    managed_by: str | None = None
    sku: str | None = None
    identity: str | None = None
    state: Any = None


@dataclass
class NsgItems:
    """attr class to build NSG rule dictionary."""

    rule_name: str | None = None
    description: str | None = None
    protocol: str | None = None
    direction: str | None = None
    src_ports: str | None = None
    dst_ports: str | None = None
    src_addrs: str | None = None
    dst_addrs: str | None = None
    action: str | None = None


@dataclass
class InterfaceItems:
    """attr class to build network interface details dictionary."""

    interface_id: str | None = None
    private_ip: str | None = None
    private_ip_allocation: str | None = None
    public_ip: str | None = None
    public_ip_allocation: str | None = None
    app_sec_group: list = field(default_factory=list)
    subnet: str | None = None
    subnet_nsg: Any = None
    subnet_route_table: Any = None


class AzureData:  # pylint:disable=too-many-instance-attributes
    """Class for returning data on an Azure tenant."""

    def __init__(
        self: AzureData,
        *,
        connect: bool = False,
        cloud: str | None = None,
    ) -> None:
        """Initialize connector for Azure Python SDK."""
        self.az_cloud_config = AzureCloudConfig(cloud)
        self.connected = False
        self.credentials: AzCredentials | None = None
        self.sub_client: SubscriptionClient | None = None
        self.resource_client: ResourceManagementClient | None = None
        self.network_client: NetworkManagementClient | None = None
        self.monitoring_client: MonitorManagementClient | None = None
        self.compute_client: ComputeManagementClient | None = None
        self.cloud: str | None = cloud or self.az_cloud_config.cloud
        self.endpoints: dict[str, Any] = self.az_cloud_config.endpoints
        self._token: str | None = None
        self.sent_urls: dict[str, Any] = {}
        self.base_url: str = ""
        self.url: str | None = None
        logger.info("Initialized AzureData")
        if connect:
            self.connect()

    def connect(
        self: Self,
        auth_methods: list[str] | None = None,
        tenant_id: str | None = None,
        *,
        silent: bool = False,
        cloud: str | None = None,
        **kwargs,
    ) -> None:
        """
        Authenticate to the Azure SDK.

        Parameters
        ----------
        auth_methods : List, optional
            list of preferred authentication methods to use, by default None
        tenant_id : str, optional
            The tenant to authenticate against. If not supplied, the default
            tenant for the identity will be used.
        silent : bool, optional
            Set true to prevent output during auth process, by default False
        cloud : str, optional
            What Azure cloud to connect to.
            By default it will attempt to use the cloud setting from config file.
            If this is not set it will default to Azure Public Cloud
        **kwargs
            Additional keyword arguments to pass to the az_connect function.

        Raises
        ------
        CloudError
            If no valid credentials are found or if subscription client can't be created

        See Also
        --------
        msticpy.auth.azure_auth.az_connect : function to authenticate to Azure SDK

        """
        if cloud:
            logger.info("Setting cloud to %s", cloud)
            self.cloud = cloud
            self.az_cloud_config = AzureCloudConfig(self.cloud)
        auth_methods = auth_methods or self.az_cloud_config.auth_methods
        tenant_id = tenant_id or self.az_cloud_config.tenant_id
        self.credentials = az_connect(
            auth_methods=auth_methods,
            tenant_id=tenant_id,
            silent=silent,
            **kwargs,
        )
        if not self.credentials:
            err_msg: str = "Could not obtain credentials."
            raise CloudError(err_msg)
        if only_interactive_cred(self.credentials.modern) and not silent:
            logger.warning("Check your default browser for interactive sign-in prompt.")

        self.sub_client = SubscriptionClient(
            credential=self.credentials.modern,
            base_url=self.az_cloud_config.resource_manager,
            credential_scopes=[self.az_cloud_config.token_uri],
        )
        if not self.sub_client:
            err_msg = "Could not create a Subscription client."
            raise CloudError(err_msg)
        logger.info("Connected to Azure Subscription Client")
        self.connected = True

    def get_subscriptions(self: Self) -> pd.DataFrame:
        """
        Get details of all subscriptions within the tenant.

        Returns
        -------
        pd.DataFrame
            Details of the subscriptions present in the users tenant.

        Raises
        ------
        MsticpyNotConnectedError
            If .connect() has not been called

        """
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        subscription_ids: list[str] = []
        display_names: list[str] = []
        states: list[str] = []
        if self.sub_client:
            try:
                sub_list: Iterable[Any] = list(
                    self.sub_client.subscriptions.list(),
                )
            except AttributeError:
                self._legacy_auth("sub_client")
                sub_list = list(self.sub_client.subscriptions.list())

            for item in sub_list:
                if item.subscription_id:
                    subscription_ids.append(item.subscription_id)
                if item.display_name:
                    display_names.append(item.display_name)
                states.append(str(item.state))

        return pd.DataFrame(
            {
                "Subscription ID": subscription_ids,
                "Display Name": display_names,
                "State": states,
            },
        )

    def get_subscription_info(self: Self, sub_id: str) -> dict:
        """
        Get information on a specific subscription.

        Parameters
        ----------
        sub_id : str
            The ID of the subscription to return details on.

        Returns
        -------
        dict
            Details on the selected subscription.

        Raises
        ------
        MsticpyNotConnectedError
            If .connect() has not been called.

        """
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )
        if not self.sub_client:
            err_msg = "sub_client must be defined to retrieve subscription info"
            raise ValueError(err_msg)

        try:
            sub: Subscription = self.sub_client.subscriptions.get(sub_id)
        except AttributeError:
            self._legacy_auth("sub_client")
            sub = self.sub_client.subscriptions.get(sub_id)
        sub_loc: dict[str, Any] | None = None
        quota_id: dict[str, Any] | None = None
        spending_limit: dict[str, Any] | None = None
        if sub.subscription_policies:
            sub_loc = sub.subscription_policies.location_placement_id
            quota_id = sub.subscription_policies.quota_id
            spending_limit = sub.subscription_policies.spending_limit

        return {
            "Subscription ID": sub.subscription_id,
            "Display Name": sub.display_name,
            "State": str(sub.state),
            "Subscription Location": sub_loc,
            "Subscription Quota": quota_id,
            "Spending Limit": spending_limit,
        }

    def list_sentinel_workspaces(self: Self, sub_id: str) -> dict[str, str]:
        """
        Return a list of Microsoft Sentinel workspaces in a Subscription.

        Parameters
        ----------
        sub_id : str
            The subscription ID to get a list of workspaces from.
            If not provided it will attempt to get sub_id from config files.

        Returns
        -------
        Dict
            A dictionary of workspace names and ids

        """
        logger.info("Finding Microsoft Sentinel Workspaces...")
        res: pd.DataFrame = self.get_resources(sub_id=sub_id)
        # handle no results
        if isinstance(res, pd.DataFrame) and not res.empty:
            sentinel: pd.DataFrame = res[
                (res["resource_type"] == "Microsoft.OperationsManagement/solutions")
                & (res["name"].str.startswith("SecurityInsights"))
            ]
            workspaces: list[str] = []
            for wrkspace in sentinel["resource_id"]:
                res_details: dict[str, Any] = self.get_resource_details(
                    sub_id=sub_id,
                    resource_id=wrkspace,
                )
                workspaces.append(res_details["properties"]["workspaceResourceId"])

            workspaces_dict: dict[str, Any] = {}
            for wrkspace in workspaces:
                name: str = wrkspace.split("/")[-1]
                workspaces_dict[name] = wrkspace
            return workspaces_dict

        logger.info("No Microsoft Sentinel workspaces in %s", sub_id)
        return {}

    # Get > List Aliases
    get_sentinel_workspaces: Callable[..., dict[str, Any]] = list_sentinel_workspaces

    def get_resources(
        self: Self,
        sub_id: str,
        rgroup: str | None = None,
        *,
        get_props: bool = False,
    ) -> pd.DataFrame:
        """
        Return details on all resources in a subscription or Resource Group.

        Parameters
        ----------
        sub_id: str
            The subscription ID to get resources for
        rgroup: str (Optional)
            The name of a Resource Group to get resources for
        get_props: bool (Optional)
            Set to True if you want to get the full properties of every resource
            Warning this may be a slow process depending on the number of resources

        Returns
        -------
        pd.DataFrame
            A dataframe of resource details

        """
        # Check if connection and client required are already present
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("resource_client", sub_id)
        if not self.resource_client:
            err_msg = "Resource client must be set to retrieve resources."
            raise ValueError(err_msg)

        resources: list[Any] = []
        if rgroup is None:
            resources.extend(iter(self.resource_client.resources.list()))
        else:
            resources.extend(
                iter(self.resource_client.resources.list_by_resource_group(rgroup)),
            )

        # Warn users about getting full properties for each resource
        if get_props:
            logger.info(
                "Collecting properties for every resource may take some time...",
            )

        resource_items: list[Any] = []

        # Get properties for each resource
        for resource in resources:
            if get_props:
                if resource.type == "Microsoft.Compute/virtualMachines":
                    state: VirtualMachineInstanceView | None = self._get_compute_state(
                        resource_id=resource.id,
                        sub_id=sub_id,
                    )
                else:
                    state = None
                try:
                    props = self.resource_client.resources.get_by_id(
                        resource.id,
                        "2019-08-01",
                    ).properties

                except CloudError:
                    props = self.resource_client.resources.get_by_id(
                        resource.id,
                        self._get_api(resource_id=resource.id, sub_id=sub_id),
                    ).properties
            else:
                props = resource.properties
                state = None

            # Parse relevant resource attributes into a dataframe and return it
            resource_details = asdict(
                Items(
                    resource.id,
                    resource.name,
                    resource.type,
                    resource.location,
                    resource.tags,
                    resource.plan,
                    props,
                    resource.kind,
                    resource.managed_by,
                    resource.sku,
                    resource.identity,
                    state,
                ),
            )
            resource_items.append(resource_details)

        return pd.DataFrame(resource_items)

    def get_resource_details(
        self: Self,
        sub_id: str,
        resource_id: str | None = None,
        resource_details: dict[str, Any] | None = None,
    ) -> dict:
        """
        Return the details of a specific Azure resource.

        Parameters
        ----------
        resource_id: str, optional
            The ID of the resource to get details on
        resource_details: dict, optional
            If ID is unknown provide the following details:
                -resource_group_name
                -resource_provider_namespace
                -resource_type
                -resource_name
                -parent_resource_path
        sub_id: str
            The ID of the subscription to get resources from

        Returns
        -------
        resource_details: dict
            The details of the requested resource

        """
        # Check if connection and client required are already present
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )
        self._check_client("resource_client", sub_id)
        if not self.resource_client:
            err_msg = "Cannot get resource details if resource client is not set."
            raise ValueError(err_msg)

        # If a resource id is provided use get_by_id to get details
        if resource_id is not None:
            try:
                resource = self.resource_client.resources.get_by_id(
                    resource_id,
                    api_version=self._get_api(resource_id=resource_id, sub_id=sub_id),
                )
            except AttributeError:
                self._legacy_auth("resource_client", sub_id)
                resource = self.resource_client.resources.get_by_id(
                    resource_id,
                    api_version=self._get_api(resource_id=resource_id, sub_id=sub_id),
                )
            if resource.type == "Microsoft.Compute/virtualMachines":
                state = self._get_compute_state(resource_id=resource_id, sub_id=sub_id)
            else:
                state = None
        # If resource details are provided use get to get details
        elif resource_details is not None:
            try:
                resource = self.resource_client.resources.get(
                    resource_details["resource_group_name"],
                    resource_details["resource_provider_namespace"],
                    resource_details["parent_resource_path"],
                    resource_details["resource_type"],
                    resource_details["resource_name"],
                    api_version=self._get_api(
                        resource_provider=(
                            resource_details["resource_provider_namespace"]
                            + "/"
                            + resource_details["resource_type"]
                        ),
                        sub_id=sub_id,
                    ),
                )
            except AttributeError:
                self._legacy_auth("resource_client", sub_id)
                resource = self.resource_client.resources.get(
                    resource_details["resource_group_name"],
                    resource_details["resource_provider_namespace"],
                    resource_details["parent_resource_path"],
                    resource_details["resource_type"],
                    resource_details["resource_name"],
                    api_version=self._get_api(
                        resource_provider=(
                            resource_details["resource_provider_namespace"]
                            + "/"
                            + resource_details["resource_type"]
                        ),
                        sub_id=sub_id,
                    ),
                )
            state = None
        else:
            err_msg = "Please provide either a resource ID or resource details"
            raise ValueError(err_msg)

        # Parse relevent details into a dictionary to return
        return asdict(
            Items(
                resource.id,
                resource.name,
                resource.type,
                resource.location,
                resource.tags,
                resource.plan,
                resource.properties,
                resource.kind,
                resource.managed_by,
                resource.sku,
                resource.identity,
                state,
            ),
        )

    @staticmethod
    def _normalize_resources(
        resource_id: str | None = None,
        resource_provider: str | None = None,
    ) -> tuple[str, str]:
        """Normalize elements depending on user input type."""
        if resource_id:
            try:
                return (
                    resource_id.split("/")[6],
                    resource_id.split("/")[7],
                )
            except IndexError as idx_err:
                err_msg = (
                    "Provided Resource ID isn't in the correct format. "
                    "It should look like:\n"
                    "/subscriptions/SUB_ID/resourceGroups/RESOURCE_GROUP/"
                    "providers/NAMESPACE/SERVICE_NAME/RESOURCE_NAME "
                )
                raise MsticpyResourceError(err_msg) from idx_err

        elif resource_provider:
            try:
                return (
                    resource_provider.split("/")[0],
                    resource_provider.split("/")[1],
                )
            except IndexError as idx_err:
                err_msg = (
                    "Provided Resource Provider isn't in the correct format.\n"
                    "It should look like: NAMESPACE/SERVICE_NAME"
                )
                raise MsticpyResourceError(err_msg) from idx_err
        else:
            err_msg = "Please provide an resource ID or resource provider namespace"
            raise ValueError(err_msg)

    def _get_api(
        self: Self,
        *,
        sub_id: str,
        resource_id: str | None = None,
        resource_provider: str | None = None,
    ) -> str:
        """
        Return the latest avaliable API version for the resource.

        Parameters
        ----------
        resource_id: str, optional
            The ID of the resources to get an API version for
        sub_id: str, optional
            The ID of the subscription to get details from
        resource_provider: str, optional
            The resource provider namespace and service to get an API version for

        Returns
        -------
        api_ver: str
            The latest available non-preview API version

        """
        # Check if connection and client required are already present
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("resource_client", sub_id)
        if not self.resource_client:
            err_msg = "Resource client must be set to get api."
            raise ValueError(err_msg)

        namespace, service = AzureData._normalize_resources(
            resource_id=resource_id,
            resource_provider=resource_provider,
        )

        # Get list of API versions for the service
        try:
            provider = self.resource_client.providers.get(namespace)
        except AttributeError:
            self._legacy_auth("resource_client", sub_id)
            provider = self.resource_client.providers.get(namespace)

        if not provider.resource_types:
            resource_types = None
        else:
            resource_types = next(
                (t for t in provider.resource_types if t.resource_type == service),
                None,
            )

        # Get first API version that isn't in preview
        if not resource_types:
            err_msg = "Resource provider not found"
            raise MsticpyResourceError(err_msg)

        api_version = [
            v for v in resource_types.api_versions if "preview" not in v.lower()
        ]
        if api_version is None or not api_version:
            api_ver = resource_types.api_versions[0]
        else:
            api_ver = api_version[0]
        return str(api_ver)

    def get_network_details(
        self: Self,
        network_id: str,
        sub_id: str,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Return details related to an Azure network interface and associated NSG.

        Parameters
        ----------
        network_id: str
            The ID of the network interface to return details on
        sub_id: str
            The subscription ID that the network interface is part of

        Returns
        -------
        details: dict
            A dictionary of items related to the network interface

        """
        # Check if connection and client required are already present
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("network_client", sub_id)
        if not self.network_client:
            err_msg = "Cannot retrieve network details if the network client is not set"
            raise ValueError(err_msg)

        # Get interface details and parse relevant elements into a dataframe
        try:
            details: NetworkInterface = self.network_client.network_interfaces.get(
                network_id.split("/")[4],
                network_id.split("/")[8],
            )
        except AttributeError:
            self._legacy_auth("network_client", sub_id)
            details = self.network_client.network_interfaces.get(
                network_id.split("/")[4],
                network_id.split("/")[8],
            )

        ips: list[dict[str, Any]] = []
        for ip_addr in details.ip_configurations or []:
            ip_details: dict[str, Any] = asdict(
                InterfaceItems(
                    interface_id=network_id,
                    private_ip=ip_addr.private_ip_address,
                    private_ip_allocation=str(ip_addr.private_ip_allocation_method),
                    public_ip=(
                        ip_addr.public_ip_address.ip_address
                        if ip_addr.public_ip_address
                        else None
                    ),
                    public_ip_allocation=(
                        ip_addr.public_ip_address.public_ip_allocation_method
                        if ip_addr.public_ip_address
                        else None
                    ),
                    app_sec_group=(
                        ip_addr.application_security_groups
                        if ip_addr.application_security_groups
                        else []
                    ),
                    subnet=ip_addr.subnet.name if ip_addr.subnet else None,
                    subnet_nsg=(
                        ip_addr.subnet.network_security_group
                        if ip_addr.subnet
                        else None
                    ),
                    subnet_route_table=(
                        ip_addr.subnet.route_table if ip_addr.subnet else None
                    ),
                ),
            )
            ips.append(ip_details)

        ip_df = pd.DataFrame(ips)

        nsg_df = pd.DataFrame()
        if (
            details.network_security_group is not None
            and details.network_security_group.id is not None
        ):
            # Get NSG details and parse relevant elements into a dataframe
            nsg_details = self.network_client.network_security_groups.get(
                details.network_security_group.id.split("/")[4],
                details.network_security_group.id.split("/")[8],
            )
            nsg_rules = []
            for nsg in nsg_details.default_security_rules:
                rules = asdict(
                    NsgItems(
                        rule_name=nsg.name,
                        description=nsg.description,
                        protocol=str(nsg.protocol),
                        direction=str(nsg.direction),
                        src_ports=nsg.source_port_range,
                        dst_ports=nsg.destination_port_range,
                        src_addrs=nsg.source_address_prefix,
                        dst_addrs=nsg.destination_address_prefix,
                        action=str(nsg.access),
                    ),
                )
                nsg_rules.append(rules)

            nsg_df = pd.DataFrame(nsg_rules)

        return ip_df, nsg_df

    def get_metrics(  # pylint: disable=too-many-locals #noqa: PLR0913
        self: Self,
        metrics: str,
        resource_id: str,
        sub_id: str,
        sample_time: str = "hour",
        start_time: int = 30,
    ) -> dict[str, pd.DataFrame]:
        """
        Return specified metrics on Azure Resource.

        Parameters
        ----------
        metrics: str
            A string list of metrics you wish to collect
            (https://docs.microsoft.com/en-us/azure/azure-monitor/platform/metrics-supported)
        resource_id: str
            The resource ID of the resource to collet the metrics from
        sub_id: str
            The subscription ID that the resource is part of
        sample_time: str (Optional)
            You can select to collect the metrics every hour of minute - default is hour
            Accepted inputs = 'hour' or 'minute'
        start_time: int (Optional)
            The number of days prior to today to collect metrics for, default is 30

        Returns
        -------
        results: dict
            A Dictionary of DataFrames containing the metrics details

        """
        if sample_time.casefold().startswith("h"):
            interval = "PT1H"
        elif sample_time.casefold().startswith("m"):
            interval = "PT1M"
        else:
            err_msg: str = "invalid value for sample_time - specify 'hour', or 'minute'"
            raise ValueError(err_msg)

        # Check if connection and client required are already present
        if self.connected is False:
            err_msg = "You need to connect to the service before using this function."
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("monitoring_client", sub_id)
        if not self.monitoring_client:
            err_msg = "Cannot get metrics if monitoring client is not set."
            raise ValueError(err_msg)

        # Get metrics in one hour chunks for the last 30 days
        start = datetime.datetime.now(tz=datetime.timezone.utc).date()
        end = start - datetime.timedelta(days=start_time)

        try:
            mon_details = self.monitoring_client.metrics.list(
                resource_id,
                timespan=f"{end}/{start}",
                interval=interval,
                metricnames=f"{metrics}",
                aggregation="Total",
            )
        except AttributeError:
            self._legacy_auth("monitoring_client", sub_id)
            mon_details = self.monitoring_client.metrics.list(
                resource_id,
                timespan=f"{end}/{start}",
                interval=interval,
                metricnames=f"{metrics}",
                aggregation="Total",
            )
        results: dict[str, Any] = {}
        # Create a dict of all the results returned
        for metric in mon_details.value:
            times: list = []
            output = []
            for time in metric.timeseries:
                for data in time.data:
                    times.append(data.time_stamp)
                    output.append(data.total)
            details = pd.DataFrame({"Time": times, "Data": output})
            details = details.replace(np.nan, 0)
            results[metric.name.value] = details
        return results

    def _get_compute_state(
        self: Self,
        resource_id: str,
        sub_id: str,
    ) -> VirtualMachineInstanceView:
        """
        Return the details on a Virtual Machine instance.

        Parameters
        ----------
        resource_id: str
            The Resource ID of the Virtual Machine
        sub_id: str
            The Subscription the Virtual Machine is part of

        Returns
        -------
        instance_details: VirtualMachineInstanceView
            The details of the Virtual Machine

        """
        if self.connected is False:
            err_msg: str = (
                "You need to connect to the service before using this function."
            )
            raise MsticpyNotConnectedError(
                err_msg,
                help_uri=MsticpyAzureConfigError.DEF_HELP_URI,
                title="Please call connect() before continuing.",
            )

        self._check_client("compute_client", sub_id)
        if not self.compute_client:
            err_msg = "Cannot provide compute state if compute_client is None."
            raise ValueError(err_msg)

        # Parse the Resource ID to extract Resource Group and Resource Name
        r_details = resource_id.split("/")
        r_group = r_details[r_details.index("resourceGroups") + 1]
        name = r_details[r_details.index("virtualMachines") + 1]

        # Get VM instance details and return them
        try:
            instance_details: VirtualMachineInstanceView = (
                self.compute_client.virtual_machines.instance_view(
                    r_group,
                    name,
                )
            )
        except AttributeError:
            self._legacy_auth("compute_client", sub_id)
            instance_details = self.compute_client.virtual_machines.instance_view(
                r_group,
                name,
            )

        return instance_details

    def _check_client(self: Self, client_name: str, sub_id: str | None = None) -> None:
        """
        Check required client is present, if not create it.

        Parameters
        ----------
        client_name : str
            The name of the client to be checked.
        sub_id : str, optional
            The subscription ID for the client to connect to, by default None

        """
        if not self.credentials:
            err_msg: str = "Credentials must be provided for _check_client to work."
            raise ValueError(err_msg)
        if getattr(self, client_name) is None:
            client: type[
                SubscriptionClient
                | ResourceManagementClient
                | NetworkManagementClient
                | MonitorManagementClient
                | ComputeManagementClient
            ] = _CLIENT_MAPPING[client_name]
            if sub_id is None:
                if issubclass(client, SubscriptionClient):
                    setattr(
                        self,
                        client_name,
                        client(
                            self.credentials.modern,
                            base_url=self.az_cloud_config.resource_manager,
                            credential_scopes=[self.az_cloud_config.token_uri],
                        ),
                    )
            else:
                setattr(
                    self,
                    client_name,
                    client(
                        self.credentials.modern,
                        subscription_id=sub_id,
                        base_url=self.az_cloud_config.resource_manager,
                        credential_scopes=[self.az_cloud_config.token_uri],
                    ),
                )

        if getattr(self, client_name) is None:
            err_msg = "Could not create client"
            raise CloudError(err_msg)

    def _legacy_auth(self: Self, client_name: str, sub_id: str | None = None) -> None:
        """
        Create client with v1 authentication token.

        Parameters
        ----------
        client_name : str
            The name of the client to be checked.
        sub_id : str, optional
            The subscription ID for the client to connect to, by default None

        """
        if not self.credentials:
            err_msg: str = (
                "Credentials must be provided for legacy authentication to work."
            )
            raise ValueError(err_msg)
        client: type[
            SubscriptionClient
            | ResourceManagementClient
            | NetworkManagementClient
            | MonitorManagementClient
            | ComputeManagementClient
        ] = _CLIENT_MAPPING[client_name]
        if sub_id is None:
            if issubclass(client, SubscriptionClient):
                setattr(
                    self,
                    client_name,
                    client(
                        self.credentials.legacy,
                        base_url=self.az_cloud_config.resource_manager,
                        credential_scopes=[self.az_cloud_config.token_uri],
                    ),
                )
        else:
            setattr(
                self,
                client_name,
                client(
                    self.credentials.legacy,
                    subscription_id=sub_id,
                    base_url=self.az_cloud_config.resource_manager,
                    credential_scopes=[self.az_cloud_config.token_uri],
                ),
            )


def get_api_headers(token: str) -> dict:
    """
    Return authorization header with current token.

    Parameters
    ----------
    token : str
        Azure auth token.

    Returns
    -------
    Dict
        A dictionary of headers to be used in API calls.

    """
    return {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
    }


def get_token(
    credential: AzCredentials,
    tenant_id: str | None = None,
    cloud: str | None = None,
) -> str:
    """
    Extract token from a azure.identity object.

    Parameters
    ----------
    credential : AzCredentials
        Azure OAuth credentials.
    tenant_id : str, optional
        The tenant to connect to if not the users home tenant.
    cloud: str, optional
        The Azure cloud to connect to.

    Returns
    -------
    str
        A token to be used in API calls.

    """
    az_cloud_config = AzureCloudConfig(cloud)
    tenant_id = tenant_id or az_cloud_config.tenant_id
    if tenant_id:
        try:
            token = credential.modern.get_token(AzureCloudConfig().token_uri)
        except ClientAuthenticationError:
            credential = fallback_devicecode_creds(cloud=cloud)
            token = credential.modern.get_token(AzureCloudConfig().token_uri)
    else:
        try:
            token = credential.modern.get_token(
                AzureCloudConfig().token_uri,
                tenant_id=tenant_id,
            )
        except ClientAuthenticationError:
            credential = fallback_devicecode_creds(cloud=cloud, tenant_id=tenant_id)
            token = credential.modern.get_token(
                AzureCloudConfig().token_uri,
                tenant_id=tenant_id,
            )

    return token.token



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Sentinel Dynamic Summary Mixin class."""
from __future__ import annotations

import logging
from functools import singledispatchmethod
from typing import TYPE_CHECKING, Any, Callable, Iterable

import httpx
from typing_extensions import Self

from msticpy.context.azure.sentinel_utils import SentinelUtilsMixin

from ..._version import VERSION
from ...common.exceptions import MsticpyAzureConnectionError, MsticpyParameterError
from ...common.pkg_config import get_config, get_http_timeout
from ...data.core.data_providers import QueryProvider
from .azure_data import get_api_headers
from .sentinel_dynamic_summary_types import (
    DynamicSummary,
    DynamicSummaryItem,
    df_to_dynamic_summary,
)

if TYPE_CHECKING:
    from datetime import datetime

    import pandas as pd


__version__ = VERSION
__author__ = "Ian Hellen"

_DYN_SUM_API_VERSION = "2023-03-01-preview"

logger: logging.Logger = logging.getLogger(__name__)


class SentinelDynamicSummaryMixin(SentinelUtilsMixin):
    """Mixin class with Sentinel Dynamic Summary integrations."""

    # expose these methods as members of the Sentinel class.
    df_to_dynamic_summary: Callable[
        ...,
        DynamicSummary,
    ] = DynamicSummary.df_to_dynamic_summary
    df_to_dynamic_summaries: Callable[
        ...,
        list[DynamicSummary],
    ] = DynamicSummary.df_to_dynamic_summaries

    @classmethod
    def new_dynamic_summary(  # pylint:disable=too-many-arguments # noqa: PLR0913
        cls: type[Self],
        summary_id: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tenant_id: str | None = None,
        azure_tenant_id: str | None = None,
        search_key: str | None = None,
        tactics: str | list[str] | None = None,
        techniques: str | list[str] | None = None,
        source_info: dict[str, Any] | None = None,
        summary_items: (
            pd.DataFrame | Iterable[DynamicSummaryItem] | list[dict[str, Any]] | None
        ) = None,
    ) -> DynamicSummary:
        """
        Return a new DynamicSummary object.

        Notes
        -----
        See the DynamicSummary class documentation for details
        of expected parameters.

        See Also
        --------
        DynamicSummary

        """
        return DynamicSummary.new_dynamic_summary(
            summary_id=summary_id,
            summary_name=name,
            summary_description=description,
            tenant_id=tenant_id,
            azure_tenant_id=azure_tenant_id,
            search_key=search_key,
            tactics=tactics,
            techniques=techniques,
            source_info=source_info,
            summary_items=summary_items,
        )

    def list_dynamic_summaries(self: Self) -> pd.DataFrame:
        """
        Return current list of Dynamic Summaries from a Sentinel workspace.

        Returns
        -------
        pd.DataFrame
            The current Dynamic Summary objects.

        """
        return self._list_items(
            item_type="dynamic_summary",
            api_version=_DYN_SUM_API_VERSION,
        )

    def get_dynamic_summary(
        self: Self,
        summary_id: str,
        *,
        summary_items: bool = False,
    ) -> DynamicSummary:
        """
        Return DynamicSummary for ID.

        Parameters
        ----------
        summary_id : str
            The ID of the Dynamic summary object.
        summary_items : bool, optional
            Use a data query to retrieve the dynamic summary along
            with summary items (data records), by default, false.

        Returns
        -------
        DynamicSummary:
            DynamicSummary object.

        Raises
        ------
        MsticpyAzureConnectionError
            If API returns an error.

        """
        if summary_items:
            if not self.sent_data_query:
                try:
                    self.sent_data_query: (
                        SentinelQueryProvider | None
                    ) = SentinelQueryProvider(
                        self.default_workspace_name,  # type: ignore[attr-defined]
                    )
                    logger.info(
                        "Created sentinel query provider for %s",
                        self.default_workspace_name,  # type: ignore[attr-defined]
                    )
                except LookupError:
                    logging.info(
                        "Unable to find default workspace."
                        "Use 'sentinel.set_default_workspace(workspace='my_ws_name' "
                        "and retry.",
                    )
            if self.sent_data_query:
                logger.info("Query dynamic summary for %s", summary_id)
                return df_to_dynamic_summary(
                    self.sent_data_query.get_dynamic_summary(summary_id),
                )

        dyn_sum_url = self.sent_urls["dynamic_summary"] + f"/{summary_id}"
        params = {"api-version": _DYN_SUM_API_VERSION}
        if not self._token:
            err_msg = "Token not found, can't get dynamic summary."
            raise ValueError(err_msg)
        response = httpx.get(
            dyn_sum_url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )
        if response.is_success:
            logger.info("Query API for summary id %s", summary_id)
            return DynamicSummary.from_json(response.json())
        logger.info(
            "API query unsuccessful - status %d, response: %s",
            response.status_code,
            response.content.decode("utf-8"),
        )
        raise MsticpyAzureConnectionError(response.json())

    def create_dynamic_summary(  # pylint:disable=too-many-arguments #noqa: PLR0913
        self: Self,
        summary: DynamicSummary | None = None,
        name: str | None = None,
        description: str | None = None,
        data: pd.DataFrame | None = None,
        *,
        summary_id: str | None = None,
        tenant_id: str | None = None,
        azure_tenant_id: str | None = None,
        search_key: str | None = None,
        tactics: str | list[str] | None = None,
        techniques: str | list[str] | None = None,
        source_info: dict[str, Any] | None = None,
    ) -> str | None:
        """
        Create a Dynamic Summary in the Sentinel Workspace.

        Parameters
        ----------
        summary : DynamicSummary
            DynamicSummary instance.
        name : str
            The name of the dynamic summary to create
        description : str
            Dynamic Summary description
        data : pd.DataFrame
            The summary data
        summary_id: str | None
            Id of the summary object
        tenant_id: str | None
            Tenant Id of the Sentinel workspace
        azure_tenant_id: str | None
            Tenant Id of the Sentinel workspace
        search_key : str, optional
            Search key for the entire summary, by default None
        tactics : Union[str, List[str], None], optional
            Relevant MITRE tactics, by default None
        techniques : Union[str, List[str], None], optional
            Relevant MITRE techniques, by default None
        source_info : str, optional
            Summary source info, by default None

        Returns
        -------
        Optional[str]
            The name/ID of the dynamic summary.

        Raises
        ------
        MsticpyAzureConnectionError
            If API returns an error.

        """
        if summary is not None:
            if not summary.summary_name:
                err_msg: str = "DynamicSummary must have unique `summary_name`."
                raise MsticpyParameterError(
                    err_msg,
                    parameters="summary_name",
                )
            return self._create_dynamic_summary(summary)
        # pylint: disable=unexpected-keyword-arg
        if not name:
            err_msg = "DynamicSummary must have unique name"
            raise MsticpyParameterError(
                err_msg,
                parameters="name",
            )
        logger.info("create_dynamic_summary %s (%s)", name, description)
        return self._create_dynamic_summary(
            name,
            description=description,
            data=data,
            summary_id=summary_id,
            tenant_id=tenant_id,
            azure_tenant_id=azure_tenant_id,
            search_key=search_key,
            tactics=tactics,
            techniques=techniques,
            source_info=source_info,
        )

    @singledispatchmethod
    def _create_dynamic_summary(
        self: Self,
        summary: DynamicSummary,
    ) -> str | None:
        """
        Create a Dynamic Summary in the Sentinel Workspace.

        Parameters
        ----------
        summary : DynamicSummary
            DynamicSummary instance.

        Returns
        -------
        Optional[str]
            The name/ID of the dynamic summary.

        Raises
        ------
        MsticpyAzureConnectionError
            If API returns an error.

        """
        self.check_connected()
        dyn_sum_url = "/".join([self.sent_urls["dynamic_summary"], summary.summary_id])

        params: dict[str, str] = {"api-version": _DYN_SUM_API_VERSION}
        if not self._token:
            err_msg: str = "Token not found, can't create dynamic summary."
            raise ValueError(err_msg)
        response: httpx.Response = httpx.put(
            dyn_sum_url,
            headers=get_api_headers(self._token),
            params=params,
            content=summary.to_json_api(),
            timeout=get_http_timeout(),
        )
        logger.info(
            "_create_dynamic_summary (DynamicSummary) status %d",
            response.status_code,
        )
        if response.is_success:
            logger.info("Dynamic summary created/updated.")
            return response.json().get("name")
        logger.warning(
            "_create_dynamic_summary (DynamicSummary) failure %s",
            response.content.decode("utf-8"),
        )
        err_msg = (
            f"Dynamic summary create/update failed with status {response.status_code}"
        )
        raise MsticpyAzureConnectionError(
            err_msg,
            "Text response:",
            response.text,
        )

    @_create_dynamic_summary.register(str)
    def _(  # pylint:disable=too-many-arguments # noqa: PLR0913
        self: Self,
        name: str,
        description: str,
        data: pd.DataFrame,
        summary_id: str | None = None,
        tenant_id: str | None = None,
        azure_tenant_id: str | None = None,
        search_key: str | None = None,
        tactics: str | list[str] | None = None,
        techniques: str | list[str] | None = None,
        source_info: dict[str, Any] | None = None,
    ) -> str | None:
        """
        Create a Dynamic Summary in the Sentinel Workspace.

        Parameters
        ----------
        name : str
            The name of the dynamic summary to create
        description : str
            Dynamic Summary description
        data : pd.DataFrame
            The summary data
        summary_id: str | None
            Id of the summary object
        tenant_id: str | None
            Tenant Id of the Sentinel workspace
        azure_tenant_id: str | None
            Tenant Id of the Sentinel workspace
        search_key : str, optional
            Search key for the entire summary, by default None
        tactics : Union[str, List[str], None], optional
            Relevant MITRE tactics, by default None
        techniques : Union[str, List[str], None], optional
            Relevant MITRE techniques, by default None
        source_info : str, optional
            Summary source info, by default None

        Returns
        -------
        Optional[str]
            The name/ID of the dynamic summary.

        Raises
        ------
        MsticpyAzureConnectionError
            If API returns an error.

        """
        self.check_connected()
        summary = DynamicSummary(
            summary_name=name,
            summary_description=description,
            summary_items=data,
            summary_id=summary_id,
            tenant_id=tenant_id,
            azure_tenant_id=azure_tenant_id,
            search_key=search_key,
            tactics=tactics,
            techniques=techniques,
            source_info=source_info,
        )
        logger.info(
            "_create_dynamic_summary (DF) rows: %d",
            len(data) if data is not None else 0,
        )
        return self.create_dynamic_summary(summary)

    def delete_dynamic_summary(
        self: Self,
        summary_id: str,
    ) -> None:
        """
        Delete the Dynamic Summary for `summary_id`.

        Parameters
        ----------
        summary_id: str, optional
            The UUID of the summary to delete.

        Raises
        ------
        MsticpyAzureConnectionError
            If the API returns an error.

        """
        self.check_connected()

        dyn_sum_url = f"{self.sent_urls['dynamic_summary']}/{summary_id}"
        params = {"api-version": _DYN_SUM_API_VERSION}
        if not self._token:
            err_msg: str = "Token not found, can't delete dynamic summary."
            raise ValueError(err_msg)
        response = httpx.delete(
            dyn_sum_url,
            headers=get_api_headers(self._token),
            params=params,
            timeout=get_http_timeout(),
        )
        logger.info(
            "delete_dynamic_summary %s - status %d",
            summary_id,
            response.status_code,
        )
        if response.is_success:
            logger.info("Dynamic summary deleted.")
            return response.json().get("name")
        logger.warning(
            "delete_dynamic_summary failure %s",
            response.content.decode("utf-8"),
        )
        err_msg = f"Dynamic summary deletion failed with status {response.status_code}"
        raise MsticpyAzureConnectionError(
            err_msg,
            "Text response:",
            response.text,
        )

    def update_dynamic_summary(  # pylint:disable=too-many-arguments # noqa:PLR0913
        self: Self,
        summary: DynamicSummary | None = None,
        summary_id: str | None = None,
        data: pd.DataFrame | None = None,
        *,
        name: str | None = None,
        description: str | None = None,
        tenant_id: str | None = None,
        azure_tenant_id: str | None = None,
        search_key: str | None = None,
        tactics: str | list[str] | None = None,
        techniques: str | list[str] | None = None,
        source_info: dict[str, Any] | None = None,
    ) -> str | None:
        """
        Update a dynamic summary in the Sentinel Workspace.

        Parameters
        ----------
        summary : DynamicSummary
            DynamicSummary instance.
        summary_id : str
            The ID of the summary to update.
        data : pd.DataFrame
            The summary data
        name : str
            The name of the dynamic summary to create
        description : str
            Dynamic Summary description
        relation_name : str, optional
            The relation name, by default None
        relation_id : str, optional
            The relation ID, by default None
        search_key : str, optional
            Search key for the entire summary, by default None
        tactics : Union[str, List[str], None], optional
            Relevant MITRE tactics, by default None
        techniques : Union[str, List[str], None], optional
            Relevant MITRE techniques, by default None
        source_info : str, optional
            Summary source info, by default None
        summary_items : Union[pd, DataFrame, Iterable[DynamicSummaryItem],
            List[Dict[str, Any]]], optional
                Collection of summary items, by default
        tenant_id: str | None
            Tenant Id of the Sentinel workspace
        azure_tenant_id: str | None
            Tenant Id of the Sentinel workspace

        Returns
        -------
        Optional[str]
            The name/ID of the dynamic summary.

        Raises
        ------
        MsticpyParameterError
            If existing summary_id not supplied.
        MsticpyAzureConnectionError
            If API returns an error.

        """
        if (summary and not summary.summary_id) or (
            data is not None and not summary_id
        ):
            err_msg: str = "You must supply a summary ID to update"
            raise MsticpyParameterError(
                err_msg,
                parameters="summary_id",
            )
        logger.info(
            "update_dynamic_summary summary %s, df %s",
            summary is not None,
            data is not None,
        )
        return self.create_dynamic_summary(
            summary=summary,
            data=data,
            name=name,
            description=description,
            summary_id=summary_id,
            tenant_id=tenant_id,
            azure_tenant_id=azure_tenant_id,
            search_key=search_key,
            tactics=tactics,
            techniques=techniques,
            source_info=source_info,
        )


class SentinelQueryProvider:
    """Class to encapsulate MS Sentinel data queries."""

    _SINGLE_SUMMARY = """
    DynamicSummary
    | where SummaryId == "{summary_id}"
    | where SummaryStatus == "Active" or SummaryDataType == "SummaryItem"
    """

    def __init__(self: SentinelQueryProvider, workspace: str) -> None:
        """Initialize Sentinel Provider."""
        workspaces = get_config("AzureSentinel.Workspaces", {})
        self.workspace_config = ""
        self.workspace_alias = ""
        for ws_name, workspace_conf in workspaces.items():
            if workspace.casefold() in [
                ws_name.casefold(),
                workspace_conf.get("WorkspaceName", "").casefold(),
            ]:
                self.workspace_config = workspace_conf
                self.workspace_alias = ws_name
                logger.info("Found workspace config %s", ws_name)
                break
        else:
            err_msg: str = f"Cannot find workspace configuration for {workspace}"
            raise LookupError(err_msg)

        self.qry_prov = QueryProvider("MSSentinel")
        self.qry_prov.connect(workspace=self.workspace_alias)

    def get_dynamic_summary(self: Self, summary_id: str) -> pd.DataFrame:
        """Retrieve dynamic summary from MS Sentinel table."""
        logger.info("Dynamic summary query for %s", summary_id)
        return self.qry_prov.MSSentinel.get_dynamic_summary_by_id(summary_id=summary_id)

    def get_dynamic_summaries(
        self: Self,
        start: datetime,
        end: datetime,
    ) -> pd.DataFrame:
        """Return dynamic summaries for date range."""
        logger.info(
            "Dynamic summary query for dynamic summaries from %s to %s",
            start.isoformat(),
            end.isoformat(),
        )
        return self.qry_prov.MSSentinel.list_dynamic_summaries(start=start, end=end)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Key Vault component edit."""
from .._version import VERSION
from .ce_simple_settings import CESimpleSettings

__version__ = VERSION
__author__ = "Ian Hellen"


class CEMsticpy(CESimpleSettings):
    """Key Vault settings edit component."""

    _DESCRIPTION = "MSTICPy Settings"
    _COMP_PATH = "msticpy"
    _HELP_TEXT = """
    Set the global parameters for MSTICPy.<br>

    This section lets you set some global parameters such as:<br>

    <ul>
    <li>Paths to query files</li>
    <li>Enabling/disabling friendly exceptions</li>
    <li>HTTP timeouts</li>
    </ul>
    Note: Proxies cannot currently be set from this interface.
    Edit the msticpyconfig.yaml file directly to set proxies
    following the guidance in the link to the documentation.<br>
    """
    _HELP_URI = {
        "MSTICPy global settings": (
            "https://msticpy.readthedocs.io/en/latest/getting_started/"
            + "msticpyconfig.html#msticpy-global-settings"
        )
    }



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query Provider additional connection methods."""
import asyncio
import logging
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import partial
from itertools import tee
from typing import Any, Dict, List, Optional, Protocol, Tuple, Union

import nest_asyncio
import pandas as pd
from tqdm.auto import tqdm

from ..._version import VERSION
from ...common.exceptions import MsticpyDataQueryError
from ...common.utility.ipython import is_ipython
from ..drivers.driver_base import DriverBase, DriverProps
from .query_source import QuerySource

__version__ = VERSION
__author__ = "Ian Hellen"

logger = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods, unnecessary-ellipsis
class QueryProviderProtocol(Protocol):
    """Protocol for required properties of QueryProvider class."""

    driver_class: Any
    _driver_kwargs: Dict[str, Any]
    _additional_connections: Dict[str, Any]
    _query_provider: DriverBase

    @staticmethod
    @abstractmethod
    def _get_query_options(
        params: Dict[str, Any], kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Return any kwargs not already in params."""


# pylint: disable=super-init-not-called
class QueryProviderConnectionsMixin(QueryProviderProtocol):
    """Mixin additional connection handling QueryProvider class."""

    @staticmethod
    @abstractmethod
    def _get_query_options(
        params: Dict[str, Any], kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Return any kwargs not already in params."""

    def exec_query(self, query: str, **kwargs) -> Union[pd.DataFrame, Any]:
        """
        Execute simple query string.

        Parameters
        ----------
        query : str
            [description]
        use_connections : Union[str, List[str]]

        Other Parameters
        ----------------
        query_options : Dict[str, Any]
            Additional options passed to query driver.
        kwargs : Dict[str, Any]
            Additional options passed to query driver.

        Returns
        -------
        Union[pd.DataFrame, Any]
            Query results - a DataFrame if successful
            or a KqlResult if unsuccessful.

        """
        query_options = kwargs.pop("query_options", {}) or kwargs
        query_source = kwargs.pop("query_source", None)

        logger.info("Executing query '%s...'", query[:40])
        logger.debug("Full query: %s", query)
        logger.debug("Query options: %s", query_options)
        if not self._additional_connections:
            return self._query_provider.query(
                query, query_source=query_source, **query_options
            )
        return self._exec_additional_connections(query, **kwargs)

    def add_connection(
        self,
        connection_str: Optional[str] = None,
        alias: Optional[str] = None,
        **kwargs,
    ):
        """
        Add an additional connection for the query provider.

        Parameters
        ----------
        connection_str : Optional[str], optional
            Connection string for the provider, by default None
        alias : Optional[str], optional
            Alias to use for the connection, by default None

        Other Parameters
        ----------------
        kwargs : Dict[str, Any]
            Other connection parameters passed to the driver.

        Notes
        -----
        Some drivers may accept types other than strings for the
        `connection_str` parameter.

        """
        # create a new instance of the driver class
        new_driver = self.driver_class(**(self._driver_kwargs))
        # connect
        new_driver.connect(connection_str=connection_str, **kwargs)
        # add to collection
        driver_key = alias or str(len(self._additional_connections))
        self._additional_connections[driver_key] = new_driver

    def list_connections(self) -> List[str]:
        """
        Return a list of current connections.

        Returns
        -------
        List[str]
            The alias and connection string for each connection.

        """
        add_connections = [
            f"{alias}: {driver.current_connection}"
            for alias, driver in self._additional_connections.items()
        ]
        return [f"Default: {self._query_provider.current_connection}", *add_connections]

    # pylint: disable=too-many-locals
    def _exec_additional_connections(self, query, **kwargs) -> pd.DataFrame:
        """
        Return results of query run query against additional connections.

        Parameters
        ----------
        query : str
            The query to execute.
        progress: bool, optional
            Show progress bar, by default True
        retry_on_error: bool, optional
            Retry failed queries, by default False
        **kwargs : Dict[str, Any]
            Additional keyword arguments to pass to the query method.

        Returns
        -------
        pd.DataFrame
            The concatenated results of the query executed against all connections.

        Notes
        -----
        This method executes the specified query against all additional connections
        added to the query provider.
        If the driver supports threading or async execution, the per-connection
        queries are executed asynchronously.
        Otherwise, the queries are executed sequentially.

        """
        progress = kwargs.pop("progress", True)
        retry = kwargs.pop("retry_on_error", False)
        # Add the initial connection
        query_tasks = {
            self._query_provider.current_connection
            or "0": partial(
                self._query_provider.query,
                query,
                **kwargs,
            )
        }
        # add the additional connections
        query_tasks.update(
            {
                name: partial(connection.query, query, **kwargs)
                for name, connection in self._additional_connections.items()
            }
        )

        logger.info("Running queries for %s connections.", len(query_tasks))
        # Run the queries threaded if supported
        if self._query_provider.get_driver_property(DriverProps.SUPPORTS_THREADING):
            logger.info("Running threaded queries.")
            event_loop = _get_event_loop()
            max_workers: int = self._query_provider.get_driver_property(
                DriverProps.MAX_PARALLEL
            )
            return event_loop.run_until_complete(
                self._exec_queries_threaded(
                    query_tasks,
                    progress,
                    retry,
                    max_workers,
                )
            )

        # standard synchronous execution
        print(f"Running query for {len(self._additional_connections)} connections.")
        return self._exec_synchronous_queries(progress, query_tasks)

    def _exec_split_query(
        self,
        split_by: str,
        query_source: QuerySource,
        query_params: Dict[str, Any],
        **kwargs,
    ) -> Union[pd.DataFrame, str, None]:
        """
        Execute a query that is split into multiple queries.

        Parameters
        ----------
        split_by : str
            The time interval to split the query by.
        query_source : QuerySource
            The query to execute.
        query_params : Dict[str, Any]
            The parameters to pass to the query.

        Other Parameters
        ----------------
        debug: bool, optional
            Return queries to be executed rather than execute them, by default False
        progress: bool, optional
            Show progress bar, by default True
        retry_on_error: bool, optional
            Retry failed queries, by default False
        **kwargs : Dict[str, Any]
            Additional keyword arguments to pass to the query method.

        Returns
        -------
        pd.DataFrame
            The concatenated results of the query executed against all connections.

        Notes
        -----
        This method executes the time-chunks of the split query.
        If the driver supports threading or async execution, the sub-queries are
        executed asynchronously. Otherwise, the queries are executed sequentially.

        """
        start = query_params.pop("start", None)
        end = query_params.pop("end", None)
        progress = kwargs.pop("progress", True)
        retry = kwargs.pop("retry_on_error", False)
        debug = kwargs.pop("debug", False)
        if not (start or end):
            print("Cannot split a query with no 'start' and 'end' parameters")
            return None

        split_queries = self._create_split_queries(
            query_source=query_source,
            query_params=query_params,
            start=start,
            end=end,
            split_by=split_by,
        )
        if debug:
            return "\n\n".join(
                f"{start}-{end}\n{query}"
                for (start, end), query in split_queries.items()
            )

        query_tasks = self._create_split_query_tasks(
            query_source, query_params, split_queries, **kwargs
        )
        # Run the queries threaded if supported
        if self._query_provider.get_driver_property(DriverProps.SUPPORTS_THREADING):
            logger.info("Running threaded queries.")
            event_loop = _get_event_loop()
            max_workers: int = self._query_provider.get_driver_property(
                DriverProps.MAX_PARALLEL
            )
            return event_loop.run_until_complete(
                self._exec_queries_threaded(
                    query_tasks,
                    progress,
                    retry,
                    max_workers,
                )
            )

        # or revert to standard synchronous execution
        return self._exec_synchronous_queries(progress, query_tasks)

    def _create_split_query_tasks(
        self,
        query_source: QuerySource,
        query_params: Dict[str, Any],
        split_queries,
        **kwargs,
    ) -> Dict[str, partial]:
        """Return dictionary of partials to execute queries."""
        # Retrieve any query options passed (other than query params)
        query_options = self._get_query_options(query_params, kwargs)
        logger.info("query_options: %s", query_options)
        logger.info("kwargs: %s", kwargs)
        if "time_span" in query_options:
            del query_options["time_span"]
        return {
            f"{start}-{end}": partial(
                self.exec_query,
                query=query_str,
                query_source=query_source,
                time_span={"start": start, "end": end},
                **query_options,
            )
            for (start, end), query_str in split_queries.items()
        }

    @staticmethod
    def _exec_synchronous_queries(
        progress: bool, query_tasks: Dict[str, Any]
    ) -> pd.DataFrame:
        logger.info("Running queries sequentially.")
        results: List[pd.DataFrame] = []
        if progress:
            query_iter = tqdm(query_tasks.items(), unit="sub-queries", desc="Running")
        else:
            query_iter = query_tasks.items()
        for con_name, query_task in query_iter:
            try:
                results.append(query_task())
            except MsticpyDataQueryError:
                print(f"Query {con_name} failed.")
        if results:
            return pd.concat(results)

        logger.warning("All queries failed.")
        return pd.DataFrame()

    def _create_split_queries(
        self,
        query_source: QuerySource,
        query_params: Dict[str, Any],
        start: datetime,
        end: datetime,
        split_by: str,
    ) -> Dict[Tuple[datetime, datetime], str]:
        """Return separate queries for split time ranges."""
        try:
            if split_by.strip().endswith("H"):
                split_by = split_by.replace("H", "h")
            split_delta = pd.Timedelta(split_by)
        except ValueError:
            split_delta = pd.Timedelta("1D")
        logger.info("Using split delta %s", split_delta)

        ranges = _calc_split_ranges(start, end, split_delta)

        split_queries = {
            (q_start, q_end): query_source.create_query(
                formatters=self._query_provider.formatters,
                start=q_start,
                end=q_end,
                **query_params,
            )
            for q_start, q_end in ranges
        }
        logger.info("Split query into %s chunks", len(split_queries))
        return split_queries

    @staticmethod
    async def _exec_queries_threaded(
        query_tasks: Dict[str, partial],
        progress: bool = True,
        retry: bool = False,
        max_workers: int = 4,
    ) -> pd.DataFrame:
        """Return results of multiple queries run as threaded tasks."""
        logger.info("Running threaded queries for %d connections.", len(query_tasks))

        event_loop = _get_event_loop()

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            # add the additional connections
            thread_tasks = {
                query_id: event_loop.run_in_executor(executor, query_func)
                for query_id, query_func in query_tasks.items()
            }
            results: List[pd.DataFrame] = []
            failed_tasks_ids: List[str] = []
            if progress:
                task_iter = tqdm(
                    asyncio.as_completed(thread_tasks.values()),
                    unit="sub-queries",
                    desc="Running",
                )
            else:
                task_iter = asyncio.as_completed(thread_tasks.values())
            ids_and_tasks = dict(zip(thread_tasks, task_iter))
            for query_id, thread_task in ids_and_tasks.items():
                try:
                    result = await thread_task
                    logger.info("Query task '%s' completed successfully.", query_id)
                    results.append(result)
                except Exception:  # pylint: disable=broad-exception-caught
                    logger.warning(
                        "Query task '%s' failed with exception",
                        query_id,
                    )
                    # Reusing thread task would result in:
                    # RuntimeError: cannot reuse already awaited coroutine
                    # A new task should be queued
                    failed_tasks_ids.append(query_id)

        # Sort the results by the order of the tasks
        results = [result for _, result in sorted(zip(thread_tasks, results))]

        if retry and failed_tasks_ids:
            failed_results: pd.DataFrame = (
                await QueryProviderConnectionsMixin._exec_queries_threaded(
                    {
                        failed_tasks_id: query_tasks[failed_tasks_id]
                        for failed_tasks_id in failed_tasks_ids
                    },
                    progress=progress,
                    retry=False,
                    max_workers=max_workers,
                )
            )
            if not failed_results.empty:
                results.append(failed_results)
        if results:
            return pd.concat(results, ignore_index=True)

        logger.warning("All queries failed.")
        return pd.DataFrame()


def _get_event_loop() -> asyncio.AbstractEventLoop:
    """Return the current event loop, or create a new one."""
    try:
        if is_ipython():
            nest_asyncio.apply()
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def _calc_split_ranges(start: datetime, end: datetime, split_delta: pd.Timedelta):
    """Return a list of time ranges split by `split_delta`."""
    # Use pandas date_range and split the result into 2 iterables
    s_ranges, e_ranges = tee(pd.date_range(start, end, freq=split_delta))
    next(e_ranges, None)  # skip to the next item in the 2nd iterable
    # Zip them together to get a list of (start, end) tuples of ranges
    # Note: we subtract 1 nanosecond from the 'end' value of each range so
    # to avoid getting duplicated records at the boundaries of the ranges.
    # Some providers don't have nanosecond granularity so we might
    # get duplicates in these cases
    ranges = [
        (s_time, e_time - pd.Timedelta("1ns"))
        for s_time, e_time in zip(s_ranges, e_ranges)
    ]

    # Since the generated time ranges are based on deltas from 'start'
    # we need to adjust the end time on the final range.
    # If the difference between the calculated last range end and
    # the query 'end' that the user requested is small (< 0.1% of a delta),
    # we just replace the last "end" time with our query end time.
    if (end - ranges[-1][1]) < (split_delta / 1000):
        ranges[-1] = ranges[-1][0], pd.Timestamp(end)
    else:
        # otherwise append a new range starting after the last range
        # in ranges and ending in 'end"
        # note - we need to add back our subtracted 1 nanosecond
        ranges.append((ranges[-1][1] + pd.Timedelta("1ns"), pd.Timestamp(end)))

    return ranges
